# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import torch

from mmagic.registry import MODELS
from mmagic.utils import tensor2img
from .base_edit_model import BaseEditModel

# TODO tensor2img will be move


@MODELS.register_module()
class BasicInterpolator(BaseEditModel):
    """Basic model for video interpolation.

    It must contain a generator that takes frames as inputs and outputs an
    interpolated frame. It also has a pixel-wise loss for training.

    Args:
        generator (dict): Config for the generator structure.
        pixel_loss (dict): Config for pixel-wise loss.
        train_cfg (dict): Config for training. Default: None.
        test_cfg (dict): Config for testing. Default: None.
        required_frames (int): Required frames in each process. Default: 2
        step_frames (int): Step size of video frame interpolation. Default: 1
        init_cfg (dict, optional): The weight initialized config for
            :class:`BaseModule`.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`.

    Attributes:
        init_cfg (dict, optional): Initialization config dict.
        data_preprocessor (:obj:`BaseDataPreprocessor`): Used for
            pre-processing data sampled by dataloader to the format accepted by
            :meth:`forward`.
    """

    def __init__(self,
                 generator: dict,
                 pixel_loss: dict,
                 train_cfg: Optional[dict] = None,
                 test_cfg: Optional[dict] = None,
                 required_frames: int = 2,
                 step_frames: int = 1,
                 init_cfg: Optional[dict] = None,
                 data_preprocessor: Optional[dict] = None):

        super().__init__(
            generator=generator,
            pixel_loss=pixel_loss,
            train_cfg=train_cfg,
            test_cfg=test_cfg,
            init_cfg=init_cfg,
            data_preprocessor=data_preprocessor)

        # Required frames in each process
        self.required_frames = required_frames
        # Step size of video frame interpolation
        self.step_frames = step_frames

    def split_frames(self, input_tensors: torch.Tensor) -> torch.Tensor:
        """split input tensors for inference.

        Args:
            input_tensors (Tensor): Tensor of input frames with shape
                [1, t, c, h, w]

        Returns:
            Tensor: Split tensor with shape [t-1, 2, c, h, w]
        """

        num_frames = input_tensors.shape[1]

        result = [
            input_tensors[:, i:i + self.required_frames]
            for i in range(0, num_frames - self.required_frames +
                           1, self.step_frames)
        ]
        result = torch.cat(result, dim=0)

        return result

    @staticmethod
    def merge_frames(input_tensors: torch.Tensor,
                     output_tensors: torch.Tensor) -> list:
        """merge input frames and output frames.

        Interpolate a frame between the given two frames.

        Merged from
            [[in1, in2], [in2, in3], [in3, in4], ...]
            [[out1], [out2], [out3], ...]
        to
            [in1, out1, in2, out2, in3, out3, in4, ...]

        Args:
            input_tensors (Tensor): The input frames with shape [n, 2, c, h, w]
            output_tensors (Tensor): The output frames with shape
                [n, 1, c, h, w].

        Returns:
            list[np.array]: The final frames.
        """

        num_frames = input_tensors.shape[0]
        result = []
        for i in range(num_frames):
            result.append(tensor2img(input_tensors[i, 0]))
            result.append(tensor2img(output_tensors[i, 0]))
        result.append(tensor2img(input_tensors[-1, 1]))

        return result



# follow pytorch GAN-Studio, random flip is used in the dataset
_base_ = [
    '../_base_/models/sngan_proj/base_sngan_proj_32x32.py',
    '../_base_/datasets/cifar10_nopad.py',
    '../_base_/gen_default_runtime.py',
]

# MODEL
discriminator_steps = 5
num_classes = 10
init_cfg = dict(type='studio')
model = dict(
    num_classes=num_classes,
    # CIFAR images are RGB, convert to BGR
    data_preprocessor=dict(output_channel_order='BGR'),
    generator=dict(
        act_cfg=dict(type='ReLU', inplace=True),
        num_classes=num_classes,
        init_cfg=init_cfg),
    discriminator=dict(
        act_cfg=dict(type='ReLU', inplace=True),
        num_classes=num_classes,
        init_cfg=init_cfg),
    discriminator_steps=discriminator_steps)

# TRAINING
train_cfg = dict(max_iters=100000 * discriminator_steps)
train_dataloader = dict(batch_size=64)  # train on 1 gpu

optim_wrapper = dict(
    generator=dict(optimizer=dict(type='Adam', lr=0.0002, betas=(0.5, 0.999))),
    discriminator=dict(
        optimizer=dict(type='Adam', lr=0.0002, betas=(0.5, 0.999))))

# VIS_HOOK
custom_hooks = [
    dict(
        type='VisualizationHook',
        interval=5000,
        fixed_input=True,
        vis_kwargs_list=dict(type='GAN', name='fake_img'))
]

# METRICS
inception_pkl = './work_dirs/inception_pkl/cifar10-full.pkl'
metrics = [
    dict(
        type='InceptionScore',
        prefix='IS-50k',
        fake_nums=50000,
        inception_style='StyleGAN',
        sample_model='orig'),
    dict(
        type='FrechetInceptionDistance',
        prefix='FID-Full-50k',
        fake_nums=50000,
        inception_style='StyleGAN',
        inception_pkl=inception_pkl,
        sample_model='orig')
]

# save multi best checkpoints
default_hooks = dict(
    checkpoint=dict(
        save_best=['FID-Full-50k/fid', 'IS-50k/is'], rule=['less', 'greater']))

# EVALUATION
val_dataloader = test_dataloader = dict(batch_size=64)
val_evaluator = test_evaluator = dict(metrics=metrics)



# Copyright (c) OpenMMLab. All rights reserved.
from logging import WARNING
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine import print_log
from mmengine.model import is_model_wrapper
from mmengine.optim import OptimWrapperDict
from PIL import Image
from torch import Tensor
from tqdm import tqdm

from mmagic.models.utils import build_module
from mmagic.registry import MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import SampleList
from ..stable_diffusion import StableDiffusion
from .controlnet_utils import change_base_model

ModelType = Union[Dict, nn.Module]


@MODELS.register_module()
class ControlStableDiffusion(StableDiffusion):
    """Implementation of `ControlNet with Stable Diffusion.

    <https://arxiv.org/abs/2302.05543>`_ (ControlNet).

    Args:
        vae (Union[dict, nn.Module]): The config or module for VAE model.
        text_encoder (Union[dict, nn.Module]): The config or module for text
            encoder.
        tokenizer (str): The **name** for CLIP tokenizer.
        unet (Union[dict, nn.Module]): The config or module for Unet model.
        controlnet (Union[dict, nn.Module]): The config or module for
            ControlNet.
        schedule (Union[dict, nn.Module]): The config or module for diffusion
            scheduler.
        test_scheduler (Union[dict, nn.Module], optional): The config or
            module for diffusion scheduler in test stage (`self.infer`). If not
            passed, will use the same scheduler as `schedule`. Defaults to
            None.
        dtype (str, optional): The dtype for the model. Defaults to 'fp16'.
        enable_xformers (bool, optional): Whether to use xformers.
            Defaults to True.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`. Defaults to
                dict(type='DataPreprocessor').
        init_cfg (dict, optional): The weight initialized config for
            :class:`BaseModule`. Defaults to None/
    """

    def __init__(self,
                 vae: ModelType,
                 text_encoder: ModelType,
                 tokenizer: str,
                 unet: ModelType,
                 controlnet: ModelType,
                 scheduler: ModelType,
                 test_scheduler: Optional[ModelType] = None,
                 dtype: str = 'fp32',
                 enable_xformers: bool = True,
                 tomesd_cfg: Optional[dict] = None,
                 data_preprocessor=dict(type='DataPreprocessor'),
                 init_cfg: Optional[dict] = None):
        super().__init__(vae, text_encoder, tokenizer, unet, scheduler,
                         test_scheduler, dtype, enable_xformers, tomesd_cfg,
                         data_preprocessor, init_cfg)

        default_args = dict()
        if dtype is not None:
            default_args['dtype'] = dtype
        self.controlnet = build_module(
            controlnet, MODELS, default_args=default_args)
        self.set_xformers(self.controlnet)

        self.vae.requires_grad_(False)
        self.text_encoder.requires_grad_(False)
        self.unet.requires_grad_(False)

    def init_weights(self):
        """Initialize the weights. Noted that this function will only be called
        at train. If you want to inference with a different unet model, you can
        call this function manually or use
        `mmagic.models.editors.controlnet.controlnet_utils.change_base_model`
        to convert the weight of ControlNet manually.

        Example:
        >>> 1. init controlnet from unet
        >>> init_cfg = dict(type='init_from_unet')

        >>> 2. switch controlnet weight from unet
        >>> # base model is not defined, use `runwayml/stable-diffusion-v1-5`
        >>> # as default
        >>> init_cfg = dict(type='convert_from_unet')
        >>> # base model is defined
        >>> init_cfg = dict(
        >>>     type='convert_from_unet',
        >>>     base_model=dict(
        >>>         type='UNet2DConditionModel',
        >>>         from_pretrained='REPO_ID',
        >>>         subfolder='unet'))
        """
        if self.init_cfg is not None:
            init_type = self.init_cfg.get('type', None)
        else:
            init_type = None

        if init_type == 'init_from_unet':
            # fetch module
            if is_model_wrapper(self.controlnet):
                controlnet = self.controlnet.module
            else:
                controlnet = self.controlnet

            if is_model_wrapper(self.unet):
                unet = self.unet.module
            else:
                unet = self.unet

            if controlnet._from_pretrained is not None:
                print_log(
                    'ControlNet has initialized from pretrained '
                    f'weight \'{controlnet._from_pretrained}\'.'
                    ' Re-initialize ControlNet from Unet.', 'current', WARNING)

            # copy weight
            log_template = 'Initialize weight ControlNet from Unet: {}'
            for n, p in unet.named_parameters():
                if n in controlnet.state_dict():
                    print_log(log_template.format(n), 'current')
                    controlnet.state_dict()[n].copy_(p.data)

            # check zero_conv
            zero_conv_blocks = controlnet.controlnet_down_blocks
            for n, p in zero_conv_blocks.named_parameters():
                if not (p == 0).all():
                    print_log(f'{n} in ControlNet is not initialized with '
                              'zero. Set to zero manually.')
                    p.data.zero_()

        elif init_type == 'convert_from_unet':
            # fetch module
            if is_model_wrapper(self.controlnet):
                controlnet = self.controlnet.module
            else:
                controlnet = self.controlnet

            if is_model_wrapper(self.unet):
                unet = self.unet.module
            else:
                unet = self.unet

            # use sd-v15 as base model by default
            base_model_default_cfg = dict(
                type='UNet2DConditionModel',
                from_pretrained='runwayml/stable-diffusion-v1-5',
                subfolder='unet')
            base_model_cfg = self.init_cfg.get('base_model',
                                               base_model_default_cfg)
            base_model = MODELS.build(base_model_cfg)
            change_base_model(controlnet, unet, base_model)

        else:
            assert init_type is None, (
                'Only support \'init_from_unet\', \'convert_from_unet\' or '
                f'None. But receive {init_type}.')

    def train_step(self, data: dict,
                   optim_wrapper: OptimWrapperDict) -> Dict[str, Tensor]:
        """Train step for ControlNet model.
        Args:
            data (dict): Data sampled from dataloader.
            optim_wrapper (OptimWrapperDict): OptimWrapperDict instance
                contains OptimWrapper of generator and discriminator.

        Returns:
            Dict[str, torch.Tensor]: A ``dict`` of tensor for logging.
        """
        data = self.data_preprocessor(data)
        inputs, data_samples = data['inputs'], data['data_samples']
        optimizer = optim_wrapper['controlnet']

        with optimizer.optim_context(self.controlnet):
            target = inputs['target']
            control = (inputs['source'] + 1) / 2  # [-1, 1] -> [0, 1]
            prompt = data_samples.prompt

            num_batches = target.shape[0]

            latents = self.vae.encode(target).latent_dist.sample()
            latents = latents * self.vae.config.scaling_factor

            noise = torch.randn_like(latents)
            timesteps = torch.randint(
                0,
                self.scheduler.num_train_timesteps, (num_batches, ),
                device=self.device)
            timesteps = timesteps.long()

            noisy_latents = self.scheduler.add_noise(latents, noise, timesteps)

            input_ids = self.tokenizer(
                prompt,
                max_length=self.tokenizer.model_max_length,
                return_tensors='pt',
                padding='max_length',
                truncation=True)['input_ids'].to(self.device)

            encoder_hidden_states = self.text_encoder(input_ids)[0]

            if self.scheduler.config.prediction_type == 'epsilon':
                gt = noise
            elif self.scheduler.config.prediction_type == 'v_prediction':
                gt = self.scheduler.get_velocity(latents, noise, timesteps)
            else:
                raise ValueError('Unknown prediction type '
                                 f'{self.scheduler.config.prediction_type}')

            # forward control
            down_block_res_samples, mid_block_res_sample = self.controlnet(
                noisy_latents,
                timesteps,
                encoder_hidden_states=encoder_hidden_states,
                controlnet_cond=control,
                return_dict=False,
            )

            # Predict the noise residual and compute loss
            model_output = self.unet(
                noisy_latents,
                timesteps,
                encoder_hidden_states=encoder_hidden_states,
                down_block_additional_residuals=down_block_res_samples,
                mid_block_additional_residual=mid_block_res_sample)
            model_pred = model_output['sample']

            loss = F.mse_loss(model_pred.float(), gt.float(), reduction='mean')

            optimizer.update_params(loss)

        return dict(loss=loss)

    def val_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Calls
        ``self.data_preprocessor`` and ``self.infer`` in order. Return the
        generated results which will be passed to evaluator or visualizer.

        Args:
            data (dict or tuple or list): Data sampled from dataset.

        Returns:
            SampleList: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        prompt = data['data_samples'].prompt
        control = data['inputs']['source']
        output = self.infer(
            prompt, control=((control + 1) / 2), return_type='tensor')

        samples = output['samples']
        samples = self.data_preprocessor.destruct(
            samples, data['data_samples'], key='target')
        control = self.data_preprocessor.destruct(
            control, data['data_samples'], key='source')

        data_sample = DataSample(
            fake_img=samples,
            control=control,
            prompt=data['data_samples'].prompt)
        data_sample_list = data_sample.split()
        return data_sample_list

    def test_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Calls
        ``self.data_preprocessor`` and ``self.infer`` in order. Return the
        generated results which will be passed to evaluator or visualizer.

        Args:
            data (dict or tuple or list): Data sampled from dataset.

        Returns:
            SampleList: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        prompt = data['data_samples'].prompt
        control = data['inputs']['source']
        output = self.infer(
            prompt, control=((control + 1) / 2), return_type='tensor')

        samples = output['samples']
        samples = self.data_preprocessor.destruct(
            samples, data['data_samples'], key='target')
        control = self.data_preprocessor.destruct(
            control, data['data_samples'], key='source')

        data_sample = DataSample(
            fake_img=samples,
            control=control,
            prompt=data['data_samples'].prompt)
        data_sample_list = data_sample.split()
        return data_sample_list

    # NOTE: maybe we should do this in a controlnet preprocessor
    @staticmethod
    def prepare_control(image: Tuple[Image.Image, List[Image.Image], Tensor,
                                     List[Tensor]], width: int, height: int,
                        batch_size: int, num_images_per_prompt: int,
                        device: str, dtype: str) -> Tensor:
        """A helper function to prepare single control images.

        Args:
            image (Tuple[Image.Image, List[Image.Image], Tensor, List[Tensor]]):  # noqa
                The input image for control.
            batch_size (int): The batch size of the control. The control will
                be repeated for `batch_size` times.
            num_images_per_prompt (int): The number images generate for one
                prompt.
            device (str): The device of the control.
            dtype (str): The dtype of the control.

        Returns:
            Tensor: The control in torch.tensor.
        """
        if not isinstance(image, torch.Tensor):
            if isinstance(image, Image.Image):
                image = [image]

            if isinstance(image[0], Image.Image):
                image = [
                    img.resize((width, height), resample=Image.LANCZOS)
                    for img in image
                ]
                image = [np.array(img)[None, :] for img in image]
                image = np.concatenate(image, axis=0)
                image = np.array(image).astype(np.float32) / 255.0
                image = image.transpose(0, 3, 1, 2)
                image = torch.from_numpy(image)
            elif isinstance(image[0], torch.Tensor):
                image = torch.cat(image, dim=0)

        image_batch_size = image.shape[0]

        if image_batch_size == 1:
            repeat_by = batch_size
        else:
            # image batch size is the same as prompt batch size
            repeat_by = num_images_per_prompt

        image = image.repeat_interleave(repeat_by, dim=0)

        image = image.to(device=device, dtype=dtype)

        return image

    @torch.no_grad()
    def infer(self,
              prompt: Union[str, List[str]],
              height: Optional[int] = None,
              width: Optional[int] = None,
              control: Optional[Union[str, np.ndarray, torch.Tensor]] = None,
              controlnet_conditioning_scale: float = 1.0,
              num_inference_steps: int = 20,
              guidance_scale: float = 7.5,
              negative_prompt: Optional[Union[str, List[str]]] = None,
              num_images_per_prompt: Optional[int] = 1,
              eta: float = 0.0,
              generator: Optional[torch.Generator] = None,
              latents: Optional[torch.FloatTensor] = None,
              return_type='image',
              show_progress=True):
        """Function invoked when calling the pipeline for generation.

        Args:
            prompt (str or List[str]): The prompt or prompts to guide
                the image generation.
            height (int, Optional): The height in pixels of the generated
                image. If not passed, the height will be
                `self.unet_sample_size * self.vae_scale_factor` Defaults
                to None.
            width (int, Optional): The width in pixels of the generated image.
                If not passed, the width will be
                `self.unet_sample_size * self.vae_scale_factor` Defaults
                to None.
            num_inference_steps (int): The number of denoising steps.
                More denoising steps usually lead to a higher quality image at
                the expense of slower inference. Defaults to 50.
            guidance_scale (float): Guidance scale as defined in Classifier-
                Free Diffusion Guidance (https://arxiv.org/abs/2207.12598).
                Defaults to 7.5
            negative_prompt (str or List[str], optional): The prompt or
                prompts not to guide the image generation. Ignored when not
                using guidance (i.e., ignored if `guidance_scale` is less
                than 1). Defaults to None.
            num_images_per_prompt (int): The number of images to generate
                per prompt. Defaults to 1.
            eta (float): Corresponds to parameter eta (η) in the DDIM paper:
                https://arxiv.org/abs/2010.02502. Only applies to
                DDIMScheduler, will be ignored for others. Defaults to 0.0.
            generator (torch.Generator, optional): A torch generator to make
                generation deterministic. Defaults to None.
            latents (torch.FloatTensor, optional): Pre-generated noisy latents,
                sampled from a Gaussian distribution, to be used as inputs for
                image generation. Can be used to tweak the same generation with
                different prompts. If not provided, a latents tensor will be
                generated by sampling using the supplied random `generator`.
                Defaults to None.
            return_type (str): The return type of the inference results.
                Supported types are 'image', 'numpy', 'tensor'. If 'image'
                is passed, a list of PIL images will be returned. If 'numpy'
                is passed, a numpy array with shape [N, C, H, W] will be
                returned, and the value range will be same as decoder's
                output range. If 'tensor' is passed, the decoder's output
                will be returned. Defaults to 'image'.

        Returns:
            dict: A dict containing the generated images and Control image.
        """
        assert return_type in ['image', 'tensor', 'numpy']

        # 0. Default height and width to unet
        height = height or self.unet_sample_size * self.vae_scale_factor
        width = width or self.unet_sample_size * self.vae_scale_factor

        # 1. Check inputs. Raise error if not correct
        self.check_inputs(prompt, height, width)

        # 2. Define call parameters
        batch_size = 1 if isinstance(prompt, str) else len(prompt)
        device = self.device
        # here `guidance_scale` is defined analog to the
        # guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf .
        # `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0

        if is_model_wrapper(self.controlnet):
            control_dtype = self.controlnet.module.dtype
        else:
            control_dtype = self.controlnet.dtype
        controls = self.prepare_control(
            control,
            width,
            height,
            batch_size,
            num_images_per_prompt,
            device,
            dtype=control_dtype)
        if do_classifier_free_guidance:
            controls = torch.cat([controls] * 2)

        # 3. Encode input prompt
        text_embeddings = self._encode_prompt(prompt, device,
                                              num_images_per_prompt,
                                              do_classifier_free_guidance,
                                              negative_prompt)

        # 4. Prepare timesteps
        # self.scheduler.set_timesteps(num_inference_steps, device=device)
        self.test_scheduler.set_timesteps(num_inference_steps)
        timesteps = self.test_scheduler.timesteps

        # 5. Prepare latent variables
        num_channels_latents = self.unet.in_channels
        latents = self.prepare_latents(
            batch_size * num_images_per_prompt,
            num_channels_latents,
            height,
            width,
            text_embeddings.dtype,
            device,
            generator,
            latents,
        )

        # 6. Prepare extra step kwargs.
        extra_step_kwargs = self.prepare_extra_step_kwargs(generator, eta)

        # 7. Denoising loop
        if show_progress:
            timesteps = tqdm(timesteps)
        for i, t in enumerate(timesteps):
            # expand the latents if we are doing classifier free guidance
            latent_model_input = torch.cat(
                [latents] * 2) if do_classifier_free_guidance else latents
            latent_model_input = self.test_scheduler.scale_model_input(
                latent_model_input, t)

            down_block_res_samples, mid_block_res_sample = self.controlnet(
                latent_model_input,
                t,
                encoder_hidden_states=text_embeddings,
                controlnet_cond=controls,
                return_dict=False,
            )

            down_block_res_samples = [
                down_block_res_sample * controlnet_conditioning_scale
                for down_block_res_sample in down_block_res_samples
            ]
            mid_block_res_sample *= controlnet_conditioning_scale

            # predict the noise residual
            noise_pred = self.unet(
                latent_model_input,
                t,
                encoder_hidden_states=text_embeddings,
                down_block_additional_residuals=down_block_res_samples,
                mid_block_additional_residual=mid_block_res_sample,
            )['sample']

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                noise_pred = noise_pred_uncond + guidance_scale * (
                    noise_pred_text - noise_pred_uncond)

                # compute the previous noisy sample x_t -> x_t-1
                latents = self.test_scheduler.step(
                    noise_pred, t, latents, **extra_step_kwargs)['prev_sample']

        # 8. Post-processing
        image = self.decode_latents(latents)

        if do_classifier_free_guidance:
            controls = torch.split(controls, controls.shape[0] // 2, dim=0)[0]

        if return_type == 'image':
            image = self.output_to_pil(image)
            controls = self.output_to_pil(controls * 2 - 1)
        elif return_type == 'numpy':
            image = image.cpu().numpy()
            controls = controls.cpu().numpy()
        else:
            assert return_type == 'tensor', (
                'Only support \'image\', \'numpy\' and \'tensor\' for '
                f'return_type, but receive {return_type}')

        return {'samples': image, 'controls': controls}

    def forward(self, *args, **kwargs):
        """forward is not implemented now."""
        raise NotImplementedError(
            'Forward is not implemented now, please use infer.')


@MODELS.register_module()
class ControlStableDiffusionImg2Img(ControlStableDiffusion):

    def _default_height_width(self, height, width, image):
        if isinstance(image, list):
            image = image[0]

        if height is None:
            if isinstance(image, Image.Image):
                height = image.height
            elif isinstance(image, torch.Tensor):
                height = image.shape[3]

            height = (height // 8) * 8  # round down to nearest multiple of 8

        if width is None:
            if isinstance(image, Image.Image):
                width = image.width
            elif isinstance(image, torch.Tensor):
                width = image.shape[2]

            width = (width // 8) * 8  # round down to nearest multiple of 8

        return height, width

    def get_timesteps(self, num_inference_steps, strength, device):
        # get the original timestep using init_timestep
        init_timestep = min(
            int(num_inference_steps * strength), num_inference_steps)

        t_start = max(num_inference_steps - init_timestep, 0)
        timesteps = self.test_scheduler.timesteps[t_start:]

        return timesteps, num_inference_steps - t_start

    def prepare_latents(self,
                        image,
                        timestep,
                        batch_size,
                        num_images_per_prompt,
                        dtype,
                        device,
                        generator=None,
                        noise=None):
        if not isinstance(image, (torch.Tensor, Image.Image, list)):
            raise ValueError(
                f'`image` has to be of type `torch.Tensor`, '
                f' `PIL.Image.Image` or list but is {type(image)}')

        image = image.to(device=device, dtype=dtype)

        batch_size = batch_size * num_images_per_prompt
        if isinstance(generator, list) and len(generator) != batch_size:
            raise ValueError(
                f'You have passed a list of generators of '
                f' length {len(generator)}, but requested an effective batch'
                f' size of {batch_size}. Make sure the batch size '
                f' matches the length of the generators.')

        if isinstance(generator, list):
            init_latents = [
                self.vae.encode(image[i:i +
                                      1]).latent_dist.sample(generator[i])
                for i in range(batch_size)
            ]
            init_latents = torch.cat(init_latents, dim=0)
        else:
            init_latents = self.vae.encode(image).latent_dist.sample(generator)

        init_latents = self.vae.config.scaling_factor * init_latents

        vae_encode_latents = init_latents

        if batch_size > init_latents.shape[0] and \
                batch_size % init_latents.shape[0] == 0:
            raise ValueError(
                f'Cannot duplicate `image` of batch size'
                f' {init_latents.shape[0]} to {batch_size} text prompts.')
        else:
            init_latents = torch.cat([init_latents], dim=0)

        shape = init_latents.shape
        if noise is None:
            noise = torch.randn(
                shape, generator=generator, device=device, dtype=dtype)

        # get latents
        init_latents = self.scheduler.add_noise(init_latents, noise, timestep)

        return init_latents, vae_encode_latents

    def prepare_latent_image(self, image, dtype):
        if isinstance(image, torch.Tensor):
            # Batch single image
            if image.ndim == 3:
                image = image.unsqueeze(0)

            image = image.to(dtype=dtype)
        else:
            # preprocess image
            if isinstance(image, (Image.Image, np.ndarray)):
                image = [image]

            if isinstance(image, list) and isinstance(image[0], Image.Image):
                image = [np.array(i.convert('RGB'))[None, :] for i in image]
                image = np.concatenate(image, axis=0)
            elif isinstance(image, list) and isinstance(image[0], np.ndarray):
                image = np.concatenate([i[None, :] for i in image], axis=0)

            image = image.transpose(0, 3, 1, 2)
            image = torch.from_numpy(image).to(dtype=dtype) / 127.5 - 1.0

        return image

    @torch.no_grad()
    def infer(self,
              prompt: Union[str, List[str]],
              latent_image: Union[torch.FloatTensor, Image.Image,
                                  List[torch.FloatTensor],
                                  List[Image.Image]] = None,
              latent_mask: torch.FloatTensor = None,
              strength: float = 1.0,
              height: Optional[int] = None,
              width: Optional[int] = None,
              control: Optional[Union[str, np.ndarray, torch.Tensor]] = None,
              controlnet_conditioning_scale: float = 1.0,
              num_inference_steps: int = 20,
              guidance_scale: float = 7.5,
              negative_prompt: Optional[Union[str, List[str]]] = None,
              num_images_per_prompt: Optional[int] = 1,
              eta: float = 0.0,
              generator: Optional[torch.Generator] = None,
              latents: Optional[torch.FloatTensor] = None,
              return_type='image',
              show_progress=True):
        """Function invoked when calling the pipeline for generation.

        Args:
            prompt (str or List[str]): The prompt or prompts to guide
                the image generation.
            height (int, Optional): The height in pixels of the generated
                image. If not passed, the height will be
                `self.unet_sample_size * self.vae_scale_factor` Defaults
                to None.
            width (int, Optional): The width in pixels of the generated image.
                If not passed, the width will be
                `self.unet_sample_size * self.vae_scale_factor` Defaults
                to None.
            num_inference_steps (int): The number of denoising steps.
                More denoising steps usually lead to a higher quality image at
                the expense of slower inference. Defaults to 50.
            guidance_scale (float): Guidance scale as defined in Classifier-
                Free Diffusion Guidance (https://arxiv.org/abs/2207.12598).
                Defaults to 7.5
            negative_prompt (str or List[str], optional): The prompt or
                prompts not to guide the image generation. Ignored when not
                using guidance (i.e., ignored if `guidance_scale` is less
                than 1). Defaults to None.
            num_images_per_prompt (int): The number of images to generate
                per prompt. Defaults to 1.
            eta (float): Corresponds to parameter eta (η) in the DDIM paper:
                https://arxiv.org/abs/2010.02502. Only applies to
                DDIMScheduler, will be ignored for others. Defaults to 0.0.
            generator (torch.Generator, optional): A torch generator to make
                generation deterministic. Defaults to None.
            latents (torch.FloatTensor, optional): Pre-generated noisy latents,
                sampled from a Gaussian distribution, to be used as inputs for
                image generation. Can be used to tweak the same generation with
                different prompts. If not provided, a latents tensor will be
                generated by sampling using the supplied random `generator`.
                Defaults to None.
            return_type (str): The return type of the inference results.
                Supported types are 'image', 'numpy', 'tensor'. If 'image'
                is passed, a list of PIL images will be returned. If 'numpy'
                is passed, a numpy array with shape [N, C, H, W] will be
                returned, and the value range will be same as decoder's
                output range. If 'tensor' is passed, the decoder's output
                will be returned. Defaults to 'image'.

        Returns:
            dict: A dict containing the generated images and Control image.
        """
        assert return_type in ['image', 'tensor', 'numpy']

        # 0. Default height and width to unet
        # height = height or self.unet_sample_size * self.vae_scale_factor
        # width = width or self.unet_sample_size * self.vae_scale_factor

        height, width = self._default_height_width(height, width, control)

        # 1. Check inputs. Raise error if not correct
        self.check_inputs(prompt, height, width)

        # 2. Define call parameters
        batch_size = 1 if isinstance(prompt, str) else len(prompt)
        device = self.device
        # here `guidance_scale` is defined analog to the
        # guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf .
        # `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0

        if is_model_wrapper(self.controlnet):
            control_dtype = self.controlnet.module.dtype
        else:
            control_dtype = self.controlnet.dtype
        controls = self.prepare_control(
            control,
            width,
            height,
            batch_size,
            num_images_per_prompt,
            device,
            dtype=control_dtype)
        if do_classifier_free_guidance:
            controls = torch.cat([controls] * 2)

        latent_image = self.prepare_latent_image(latent_image,
                                                 self.controlnet.dtype)

        # 3. Encode input prompt
        text_embeddings = self._encode_prompt(prompt, device,
                                              num_images_per_prompt,
                                              do_classifier_free_guidance,
                                              negative_prompt)

        # 4. Prepare timesteps
        # self.scheduler.set_timesteps(num_inference_steps, device=device)
        self.test_scheduler.set_timesteps(num_inference_steps)
        timesteps = self.test_scheduler.timesteps

        timesteps, num_inference_steps = self.get_timesteps(
            num_inference_steps, strength, device)

        latent_timestep = timesteps[:1].repeat(batch_size *
                                               num_images_per_prompt)

        # 5. Prepare latent variables
        latents, vae_encode_latents = self.prepare_latents(
            latent_image,
            latent_timestep,
            batch_size,
            num_images_per_prompt,
            text_embeddings.dtype,
            device,
            generator,
            noise=latents)

        # 6. Prepare extra step kwargs.
        extra_step_kwargs = self.prepare_test_scheduler_extra_step_kwargs(
            generator, eta)

        # 7. Denoising loop
        if show_progress:
            timesteps = tqdm(timesteps)
        for i, t in enumerate(timesteps):
            # expand the latents if we are doing classifier free guidance
            latent_model_input = torch.cat(
                [latents] * 2) if do_classifier_free_guidance else latents
            latent_model_input = self.test_scheduler.scale_model_input(
                latent_model_input, t)

            down_block_res_samples, mid_block_res_sample = self.controlnet(
                latent_model_input,
                t,
                encoder_hidden_states=text_embeddings,
                controlnet_cond=controls,
                return_dict=False,
            )

            down_block_res_samples = [
                down_block_res_sample * controlnet_conditioning_scale
                for down_block_res_sample in down_block_res_samples
            ]
            mid_block_res_sample *= controlnet_conditioning_scale

            # predict the noise residual
            noise_pred = self.unet(
                latent_model_input,
                t,
                encoder_hidden_states=text_embeddings,
                down_block_additional_residuals=down_block_res_samples,
                mid_block_additional_residual=mid_block_res_sample,
            )['sample']

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                noise_pred = noise_pred_uncond + guidance_scale * (
                    noise_pred_text - noise_pred_uncond)

                # compute the previous noisy sample x_t -> x_t-1
                latents = self.test_scheduler.step(
                    noise_pred, t, latents, **extra_step_kwargs)['prev_sample']

        if latent_mask is not None:
            latents = latents * latent_mask + \
                vae_encode_latents * (1.0 - latent_mask)

        # 8. Post-processing
        image = self.decode_latents(latents)

        if do_classifier_free_guidance:
            controls = torch.split(controls, controls.shape[0] // 2, dim=0)[0]

        if return_type == 'image':
            image = self.output_to_pil(image)
            controls = self.output_to_pil(controls * 2 - 1)
        elif return_type == 'numpy':
            image = image.cpu().numpy()
            controls = controls.cpu().numpy()
        else:
            assert return_type == 'tensor', (
                'Only support \'image\', \'numpy\' and \'tensor\' for '
                f'return_type, but receive {return_type}')

        return {'samples': image, 'controls': controls}



_base_ = [
    '../_base_/models/dcgan/base_dcgan_128x128.py',
    '../_base_/datasets/unconditional_imgs_128x128.py',
    '../_base_/gen_default_runtime.py'
]
model = dict(type='LSGAN', discriminator=dict(output_scale=4, out_channels=1))

total_iters = 160000
disc_step = 1
train_cfg = dict(max_iters=total_iters * disc_step)

# define dataset
# you must set `samples_per_gpu` and `imgs_root`
# `batch_size` and `data_root` need to be set.
batch_size = 64
data_root = './data/celeba-cropped/cropped_images_aligned_png/'

train_dataloader = dict(
    batch_size=batch_size, dataset=dict(data_root=data_root))

val_dataloader = dict(batch_size=batch_size, dataset=dict(data_root=data_root))

test_dataloader = dict(
    batch_size=batch_size, dataset=dict(data_root=data_root))

optim_wrapper = dict(
    generator=dict(optimizer=dict(type='Adam', lr=0.0001, betas=(0.5, 0.99))),
    discriminator=dict(
        optimizer=dict(type='Adam', lr=0.0001, betas=(0.5, 0.99))))

default_hooks = dict(
    checkpoint=dict(
        save_best=['FID-Full-50k/fid', 'IS-50k/is'], rule=['less', 'greater']))

# METRICS
metrics = [
    dict(
        type='InceptionScore',
        prefix='IS-50k',
        fake_nums=50000,
        inception_style='StyleGAN',
        sample_model='orig'),
    dict(
        type='FrechetInceptionDistance',
        prefix='FID-Full-50k',
        fake_nums=50000,
        inception_style='StyleGAN',
        sample_model='orig')
]

val_evaluator = dict(metrics=metrics)
test_evaluator = dict(metrics=metrics)

# TODO
# metrics = dict(
#     ms_ssim10k=dict(type='MS_SSIM', num_images=10000),
#     swd16k=dict(type='SWD', num_images=16384, image_shape=(3, 128, 128)),
#     fid50k=dict(type='FID', num_images=50000, inception_pkl=None))



# Copyright (c) OpenMMLab. All rights reserved.
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule

from mmagic.models.archs import SimpleGatedConvModule
from mmagic.registry import MODELS


@MODELS.register_module()
class DeepFillEncoder(BaseModule):
    """Encoder used in DeepFill model.

    This implementation follows:
    Generative Image Inpainting with Contextual Attention

    Args:
        in_channels (int): The number of input channels. Default: 5.
        conv_type (str): The type of conv module. In DeepFillv1 model, the
            `conv_type` should be 'conv'. In DeepFillv2 model, the `conv_type`
            should be 'gated_conv'.
        norm_cfg (dict): Config dict to build norm layer. Default: None.
        act_cfg (dict): Config dict for activation layer, "elu" by default.
        encoder_type (str): Type of the encoder. Should be one of ['stage1',
            'stage2_conv', 'stage2_attention']. Default: 'stage1'.
        channel_factor (float): The scale factor for channel size.
            Default: 1.
        kwargs (keyword arguments).
    """
    _conv_type = dict(conv=ConvModule, gated_conv=SimpleGatedConvModule)

    def __init__(self,
                 in_channels=5,
                 conv_type='conv',
                 norm_cfg=None,
                 act_cfg=dict(type='ELU'),
                 encoder_type='stage1',
                 channel_factor=1.,
                 **kwargs):
        super().__init__()
        conv_module = self._conv_type[conv_type]
        channel_list_dict = dict(
            stage1=[32, 64, 64, 128, 128, 128],
            stage2_conv=[32, 32, 64, 64, 128, 128],
            stage2_attention=[32, 32, 64, 128, 128, 128])
        channel_list = channel_list_dict[encoder_type]
        channel_list = [int(x * channel_factor) for x in channel_list]
        kernel_size_list = [5, 3, 3, 3, 3, 3]
        stride_list = [1, 2, 1, 2, 1, 1]
        for i in range(6):
            ks = kernel_size_list[i]
            padding = (ks - 1) // 2
            self.add_module(
                f'enc{i + 1}',
                conv_module(
                    in_channels,
                    channel_list[i],
                    kernel_size=ks,
                    stride=stride_list[i],
                    padding=padding,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg,
                    **kwargs))
            in_channels = channel_list[i]

    def forward(self, x):
        """Forward Function.

        Args:
            x (torch.Tensor): Input tensor with shape of (n, c, h, w).

        Returns:
            torch.Tensor: Output tensor with shape of (n, c, h', w').
        """
        for i in range(6):
            x = getattr(self, f'enc{i + 1}')(x)
        outputs = dict(out=x)
        return outputs



test_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(type='PackInputs')
]

sidd_data_root = 'data/SIDD'
sidd_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='SIDD', task_name='denoising'),
        data_root=sidd_data_root,
        data_prefix=dict(img='input', gt='groundtruth'),
        pipeline=test_pipeline))
sidd_evaluator = [
    dict(type='PSNR', prefix='SIDD'),
    dict(type='SSIM', prefix='SIDD'),
]

dnd_data_root = 'data/DND'
dnd_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='DND', task_name='denoising'),
        data_root=dnd_data_root,
        data_prefix=dict(img='input', gt='groundtruth'),
        pipeline=test_pipeline))
dnd_evaluator = [
    dict(type='PSNR', prefix='DND'),
    dict(type='SSIM', prefix='DND'),
]

# test config
test_cfg = dict(type='MultiTestLoop')
test_dataloader = [
    sidd_dataloader,
    # dnd_dataloader,
]
test_evaluator = [
    sidd_evaluator,
    # dnd_dataloader,
]



_base_ = ['./adm_ddim250_8xb32_imagenet-64x64.py']

model = dict(
    classifier=dict(
        type='EncoderUNetModel',
        image_size=64,
        in_channels=3,
        model_channels=128,
        out_channels=1000,
        num_res_blocks=4,
        attention_resolutions=(2, 4, 8),
        channel_mult=(1, 2, 3, 4),
        use_fp16=False,
        num_head_channels=64,
        use_scale_shift_norm=True,
        resblock_updown=True,
        pool='attention'))

metrics = [
    dict(
        type='FrechetInceptionDistance',
        prefix='FID-Full-50k',
        fake_nums=50000,
        inception_style='StyleGAN',
        sample_model='orig',
        sample_kwargs=dict(
            num_inference_steps=250, show_progress=True, classifier_scale=1.))
]

val_evaluator = dict(metrics=metrics)
test_evaluator = dict(metrics=metrics)



# Copyright (c) OpenMMLab. All rights reserved.
import math

import torch.nn as nn

from .swinir_utils import to_2tuple


class PatchEmbed(nn.Module):
    r""" Image to Patch Embedding
    Args:
        img_size (int): Image size.  Default: 224.
        patch_size (int): Patch token size. Default: 4.
        in_chans (int): Number of input image channels. Default: 3.
        embed_dim (int): Number of linear projection output channels.
            Default: 96.
        norm_layer (nn.Module, optional): Normalization layer. Default: None
    """

    def __init__(self,
                 img_size=224,
                 patch_size=4,
                 in_chans=3,
                 embed_dim=96,
                 norm_layer=None):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        patches_resolution = [
            img_size[0] // patch_size[0], img_size[1] // patch_size[1]
        ]
        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]

        self.in_chans = in_chans
        self.embed_dim = embed_dim

        if norm_layer is not None:
            self.norm = norm_layer(embed_dim)
        else:
            self.norm = None

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (B, C, Ph, Pw).

        Returns:
            Tensor: Forward results.
        """
        x = x.flatten(2).transpose(1, 2)  # B Ph*Pw C
        if self.norm is not None:
            x = self.norm(x)
        return x


class PatchUnEmbed(nn.Module):
    r""" Image to Patch Unembedding
    Args:
        img_size (int): Image size.  Default: 224.
        patch_size (int): Patch token size. Default: 4.
        in_chans (int): Number of input image channels. Default: 3.
        embed_dim (int): Number of linear projection output channels.
            Default: 96.
        norm_layer (nn.Module, optional): Normalization layer. Default: None
    """

    def __init__(self,
                 img_size=224,
                 patch_size=4,
                 in_chans=3,
                 embed_dim=96,
                 norm_layer=None):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        patches_resolution = [
            img_size[0] // patch_size[0], img_size[1] // patch_size[1]
        ]
        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]

        self.in_chans = in_chans
        self.embed_dim = embed_dim

    def forward(self, x, x_size):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (B, L, C).
            x_size (tuple[int]): Resolution of input feature.

        Returns:
            Tensor: Forward results.
        """
        B, HW, C = x.shape
        x = x.transpose(1, 2).view(B, self.embed_dim, x_size[0],
                                   x_size[1])  # B Ph*Pw C
        return x


class Upsample(nn.Sequential):
    """Upsample module.

    Args:
        scale (int): Scale factor. Supported scales: 2^n and 3.
        num_feat (int): Channel number of intermediate features.
    """

    def __init__(self, scale, num_feat):
        m = []
        if (scale & (scale - 1)) == 0:  # scale = 2^n
            for _ in range(int(math.log(scale, 2))):
                m.append(nn.Conv2d(num_feat, 4 * num_feat, 3, 1, 1))
                m.append(nn.PixelShuffle(2))
        elif scale == 3:
            m.append(nn.Conv2d(num_feat, 9 * num_feat, 3, 1, 1))
            m.append(nn.PixelShuffle(3))
        else:
            raise ValueError(f'scale {scale} is not supported. '
                             'Supported scales: 2^n and 3.')
        super(Upsample, self).__init__(*m)


class UpsampleOneStep(nn.Sequential):
    """UpsampleOneStep module (the difference with Upsample is that it always
    only has 1conv + 1pixelshuffle) Used in lightweight SR to save parameters.

    Args:
        scale (int): Scale factor. Supported scales: 2^n and 3.
        num_feat (int): Channel number of intermediate features.
        num_out_ch (int): Channel number for PixelShuffle.
        input_resolution (tuple[int], optional): Input resolution.
            Default: None
    """

    def __init__(self, scale, num_feat, num_out_ch, input_resolution=None):
        self.num_feat = num_feat
        self.input_resolution = input_resolution
        m = []
        m.append(nn.Conv2d(num_feat, (scale**2) * num_out_ch, 3, 1, 1))
        m.append(nn.PixelShuffle(scale))
        super(UpsampleOneStep, self).__init__(*m)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional, Sequence

import torch.nn as nn
import torch.utils.checkpoint as cp
from mmcv.cnn import build_activation_layer, build_conv_layer, build_norm_layer
from mmengine import MMLogger
from mmengine.model.weight_init import constant_init, kaiming_init
from mmengine.runner import load_checkpoint
from mmengine.utils.dl_utils.parrots_wrapper import _BatchNorm
from torch import Tensor


class BasicBlock(nn.Module):
    """Basic block for ResNet.

    Args:
        inplanes (int): Number of input channels.
        planes (int): Number of output channels.
        stride (int): Stride of the first block of one stage. Default: 1.
        dilation (int): Dilation of one stage. Default: 1.
        downsample (nn.Module): Downsample module. Default: None.
        act_cfg (dict): Dictionary to construct and config activation layer.
            Default: dict(type='ReLU').
        conv_cfg (dict): Dictionary to construct and config convolution layer.
            Default: None.
        norm_cfg (dict): Dictionary to construct and config norm layer.
            Default: dict(type='BN').
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
    """

    expansion = 1

    def __init__(self,
                 inplanes: int,
                 planes: int,
                 stride: int = 1,
                 dilation: int = 1,
                 downsample: Optional[nn.Module] = None,
                 act_cfg: dict = dict(type='ReLU'),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: dict = dict(type='BN'),
                 with_cp: bool = False):
        super(BasicBlock, self).__init__()

        self.norm1_name, norm1 = build_norm_layer(norm_cfg, planes, postfix=1)
        self.norm2_name, norm2 = build_norm_layer(norm_cfg, planes, postfix=2)

        self.conv1 = build_conv_layer(
            conv_cfg,
            inplanes,
            planes,
            3,
            stride=stride,
            padding=dilation,
            dilation=dilation,
            bias=False)
        self.add_module(self.norm1_name, norm1)
        self.conv2 = build_conv_layer(
            conv_cfg, planes, planes, 3, padding=1, bias=False)
        self.add_module(self.norm2_name, norm2)

        self.activate = build_activation_layer(act_cfg)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation
        self.with_cp = with_cp

    @property
    def norm1(self) -> nn.Module:
        """nn.Module: normalization layer after the first convolution layer"""
        return getattr(self, self.norm1_name)

    @property
    def norm2(self) -> nn.Module:
        """nn.Module: normalization layer after the second convolution layer"""
        return getattr(self, self.norm2_name)

    def forward(self, x: Tensor) -> Tensor:
        """Forward function."""

        def _inner_forward(x: Tensor) -> Tensor:
            identity = x

            out = self.conv1(x)
            out = self.norm1(out)
            out = self.activate(out)

            out = self.conv2(out)
            out = self.norm2(out)

            if self.downsample is not None:
                identity = self.downsample(x)

            out += identity

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        out = self.activate(out)

        return out


class Bottleneck(nn.Module):
    """Bottleneck block for ResNet.

    Args:
        inplanes (int): Number of input channels.
        planes (int): Number of output channels.
        stride (int): Stride of the first block of one stage. Default: 1.
        dilation (int): Dilation of one stage. Default: 1.
        downsample (nn.Module): Downsample module. Default: None.
        act_cfg (dict): Dictionary to construct and config activation layer.
            Default: dict(type='ReLU').
        conv_cfg (dict): Dictionary to construct and config convolution layer.
            Default: None.
        norm_cfg (dict): Dictionary to construct and config norm layer.
            Default: dict(type='BN').
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
    """

    expansion = 4

    def __init__(self,
                 inplanes: int,
                 planes: int,
                 stride: int = 1,
                 dilation: int = 1,
                 downsample: Optional[nn.Module] = None,
                 act_cfg: dict = dict(type='ReLU'),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: dict = dict(type='BN'),
                 with_cp: bool = False):
        super(Bottleneck, self).__init__()

        self.inplanes = inplanes
        self.planes = planes
        self.stride = stride
        self.dilation = dilation
        self.act_cfg = act_cfg
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.conv1_stride = 1
        self.conv2_stride = stride
        self.with_cp = with_cp

        self.norm1_name, norm1 = build_norm_layer(norm_cfg, planes, postfix=1)
        self.norm2_name, norm2 = build_norm_layer(norm_cfg, planes, postfix=2)
        self.norm3_name, norm3 = build_norm_layer(
            norm_cfg, planes * self.expansion, postfix=3)

        self.conv1 = build_conv_layer(
            conv_cfg,
            inplanes,
            planes,
            kernel_size=1,
            stride=self.conv1_stride,
            bias=False)
        self.add_module(self.norm1_name, norm1)

        self.conv2 = build_conv_layer(
            conv_cfg,
            planes,
            planes,
            kernel_size=3,
            stride=self.conv2_stride,
            padding=dilation,
            dilation=dilation,
            bias=False)

        self.add_module(self.norm2_name, norm2)
        self.conv3 = build_conv_layer(
            conv_cfg,
            planes,
            planes * self.expansion,
            kernel_size=1,
            bias=False)
        self.add_module(self.norm3_name, norm3)

        self.activate = build_activation_layer(act_cfg)
        self.downsample = downsample

    @property
    def norm1(self) -> nn.Module:
        """nn.Module: normalization layer after the first convolution layer"""
        return getattr(self, self.norm1_name)

    @property
    def norm2(self) -> nn.Module:
        """nn.Module: normalization layer after the second convolution layer"""
        return getattr(self, self.norm2_name)

    @property
    def norm3(self) -> nn.Module:
        """nn.Module: normalization layer after the second convolution layer"""
        return getattr(self, self.norm3_name)

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.norm1(out)
        out = self.activate(out)

        out = self.conv2(out)
        out = self.norm2(out)
        out = self.activate(out)

        out = self.conv3(out)
        out = self.norm3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.activate(out)

        return out


class ResNet(nn.Module):
    """General ResNet.

    This class is adopted from
    https://github.com/open-mmlab/mmsegmentation/blob/master/mmseg/models/backbones/resnet.py.

    Args:
        depth (int): Depth of resnet, from {18, 34, 50, 101, 152}.
        in_channels (int): Number of input image channels. Default" 3.
        stem_channels (int): Number of stem channels. Default: 64.
        base_channels (int): Number of base channels of res layer. Default: 64.
        num_stages (int): Resnet stages, normally 4.
        strides (Sequence[int]): Strides of the first block of each stage.
            Default: (1, 2, 2, 2).
        dilations (Sequence[int]): Dilation of each stage.
            Default: (1, 1, 2, 4).
        deep_stem (bool): Replace 7x7 conv in input stem with 3 3x3 conv.
            Default: False.
        avg_down (bool): Use AvgPool instead of stride conv when
            downsampling in the bottleneck. Default: False.
        frozen_stages (int): Stages to be frozen (stop grad and set eval mode).
            -1 means not freezing any parameters. Default: -1.
        act_cfg (dict): Dictionary to construct and config activation layer.
            Default: dict(type='ReLU').
        conv_cfg (dict): Dictionary to construct and config convolution layer.
            Default: None.
        norm_cfg (dict): Dictionary to construct and config norm layer.
            Default: dict(type='BN').
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        multi_grid (Sequence[int]|None): Multi grid dilation rates of last
            stage. Default: None.
        contract_dilation (bool): Whether contract first dilation of each layer
            Default: False.
        zero_init_residual (bool): Whether to use zero init for last norm layer
            in resblocks to let them behave as identity. Default: True.
    """

    arch_settings = {
        18: (BasicBlock, (2, 2, 2, 2)),
        34: (BasicBlock, (3, 4, 6, 3)),
        50: (Bottleneck, (3, 4, 6, 3)),
        101: (Bottleneck, (3, 4, 23, 3)),
        152: (Bottleneck, (3, 8, 36, 3))
    }

    def __init__(self,
                 depth: int,
                 in_channels: int = 3,
                 stem_channels: int = 64,
                 base_channels: int = 64,
                 num_stages: int = 4,
                 strides: Sequence[int] = (1, 2, 2, 2),
                 dilations: Sequence[int] = (1, 1, 2, 4),
                 deep_stem: bool = False,
                 avg_down: bool = False,
                 frozen_stages: int = -1,
                 act_cfg: dict = dict(type='ReLU'),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: dict = dict(type='BN'),
                 with_cp: bool = False,
                 multi_grid: Optional[Sequence[int]] = None,
                 contract_dilation: bool = False,
                 zero_init_residual: bool = True):
        super(ResNet, self).__init__()
        from functools import partial

        if depth not in self.arch_settings:
            raise KeyError(f'invalid depth {depth} for resnet')
        self.block, stage_blocks = self.arch_settings[depth]
        self.depth = depth
        self.inplanes = stem_channels
        self.stem_channels = stem_channels
        self.base_channels = base_channels
        self.num_stages = num_stages
        assert num_stages >= 1 and num_stages <= 4

        self.strides = strides
        self.dilations = dilations
        assert len(strides) == len(dilations) == num_stages

        self.deep_stem = deep_stem
        self.avg_down = avg_down
        self.frozen_stages = frozen_stages

        self.conv_cfg = conv_cfg
        self.act_cfg = act_cfg
        self.norm_cfg = norm_cfg

        self.with_cp = with_cp
        self.multi_grid = multi_grid
        self.contract_dilation = contract_dilation
        self.zero_init_residual = zero_init_residual

        self._make_stem_layer(in_channels, stem_channels)

        self.layer1 = self._make_layer(
            self.block, 64, stage_blocks[0], stride=strides[0])
        self.layer2 = self._make_layer(
            self.block, 128, stage_blocks[1], stride=strides[1])
        self.layer3 = self._make_layer(
            self.block, 256, stage_blocks[2], stride=strides[2])
        self.layer4 = self._make_layer(
            self.block, 512, stage_blocks[3], stride=strides[3])

        self.layer1.apply(partial(self._nostride_dilate, dilate=dilations[0]))
        self.layer2.apply(partial(self._nostride_dilate, dilate=dilations[1]))
        self.layer3.apply(partial(self._nostride_dilate, dilate=dilations[2]))
        self.layer4.apply(partial(self._nostride_dilate, dilate=dilations[3]))

        self._freeze_stages()

    def _make_stem_layer(self, in_channels: int, stem_channels: int) -> None:
        """Make stem layer for ResNet."""
        if self.deep_stem:
            self.stem = nn.Sequential(
                build_conv_layer(
                    self.conv_cfg,
                    in_channels,
                    stem_channels // 2,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=False),
                build_norm_layer(self.norm_cfg, stem_channels // 2)[1],
                build_activation_layer(self.act_cfg),
                build_conv_layer(
                    self.conv_cfg,
                    stem_channels // 2,
                    stem_channels // 2,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False),
                build_norm_layer(self.norm_cfg, stem_channels // 2)[1],
                build_activation_layer(self.act_cfg),
                build_conv_layer(
                    self.conv_cfg,
                    stem_channels // 2,
                    stem_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False),
                build_norm_layer(self.norm_cfg, stem_channels)[1],
                build_activation_layer(self.act_cfg))
        else:
            self.conv1 = build_conv_layer(
                self.conv_cfg,
                in_channels,
                stem_channels,
                kernel_size=7,
                stride=2,
                padding=3,
                bias=False)
            self.norm1_name, norm1 = build_norm_layer(
                self.norm_cfg, stem_channels, postfix=1)
            self.add_module(self.norm1_name, norm1)
            self.activate = build_activation_layer(self.act_cfg)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

    @property
    def norm1(self) -> nn.Module:
        """nn.Module: normalization layer after the second convolution layer"""
        return getattr(self, self.norm1_name)

    def _make_layer(self,
                    block: BasicBlock,
                    planes: int,
                    blocks: int,
                    stride: int = 1,
                    dilation: int = 1) -> nn.Module:
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                build_conv_layer(
                    self.conv_cfg,
                    self.inplanes,
                    planes * block.expansion,
                    stride=stride,
                    kernel_size=1,
                    dilation=dilation,
                    bias=False),
                build_norm_layer(self.norm_cfg, planes * block.expansion)[1])

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample=downsample,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
                conv_cfg=self.conv_cfg))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg,
                    conv_cfg=self.conv_cfg))

        return nn.Sequential(*layers)

    def _nostride_dilate(self, m: nn.Module, dilate: int) -> None:
        classname = m.__class__.__name__
        if classname.find('Conv') != -1 and dilate > 1:
            # the convolution with stride

            if m.stride == (2, 2):
                m.stride = (1, 1)
                if m.kernel_size == (3, 3):
                    m.dilation = (dilate // 2, dilate // 2)
                    m.padding = (dilate // 2, dilate // 2)
            # other convolutions
            else:
                if m.kernel_size == (3, 3):
                    m.dilation = (dilate, dilate)
                    m.padding = (dilate, dilate)

    def init_weights(self, pretrained: Optional[str] = None) -> None:
        """Init weights for the model.

        Args:
            pretrained (str, optional): Path for pretrained weights. If given
                None, pretrained weights will not be loaded. Defaults to None.
        """
        if isinstance(pretrained, str):
            logger = MMLogger.get_current_instance()
            load_checkpoint(self, pretrained, strict=False, logger=logger)
        elif pretrained is None:
            for m in self.modules():
                if isinstance(m, nn.Conv2d):
                    kaiming_init(m)
                elif isinstance(m, (_BatchNorm, nn.GroupNorm)):
                    constant_init(m, 1)
            if self.zero_init_residual:

                for m in self.modules():
                    if isinstance(m, Bottleneck):
                        constant_init(m.norm3, 0)
                    elif isinstance(m, BasicBlock):
                        constant_init(m.norm2, 0)
        else:
            raise TypeError('pretrained must be a str or None')

    def _freeze_stages(self) -> None:
        """Freeze stages param and norm stats."""
        if self.frozen_stages >= 0:
            if self.deep_stem:
                self.stem.eval()
                for param in self.stem.parameters():
                    param.requires_grad = False
            else:
                self.norm1.eval()
                for m in [self.conv1, self.norm1]:
                    for param in m.parameters():
                        param.requires_grad = False

        for i in range(1, self.frozen_stages + 1):
            m = getattr(self, f'layer{i}')
            m.eval()
            for param in m.parameters():
                param.requires_grad = False

    def forward(self, x: Tensor) -> List[Tensor]:
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (N, C, H, W).

        Returns:
            Tensor: Output tensor.
        """
        conv_out = [x]
        if self.deep_stem:
            x = self.stem(x)
        else:
            x = self.conv1(x)
            x = self.norm1(x)
            x = self.activate(x)
        conv_out.append(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        conv_out.append(x)
        x = self.layer2(x)
        conv_out.append(x)
        x = self.layer3(x)
        conv_out.append(x)
        x = self.layer4(x)
        conv_out.append(x)
        return conv_out



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import numpy as np

from mmagic.registry import METRICS
from .base_sample_wise_metric import BaseSampleWiseMetric
from .metrics_utils import img_transform


@METRICS.register_module()
class SNR(BaseSampleWiseMetric):
    """Signal-to-Noise Ratio.

    Ref: https://en.wikipedia.org/wiki/Signal-to-noise_ratio

    Args:

        gt_key (str): Key of ground-truth. Default: 'gt_img'
        pred_key (str): Key of prediction. Default: 'pred_img'
        collect_device (str): Device name used for collecting results from
            different ranks during distributed training. Must be 'cpu' or
            'gpu'. Defaults to 'cpu'.
        prefix (str, optional): The prefix that will be added in the metric
            names to disambiguate homonymous metrics of different evaluators.
            If prefix is not provided in the argument, self.default_prefix
            will be used instead. Default: None
        crop_border (int): Cropped pixels in each edges of an image. These
            pixels are not involved in the SNR calculation. Default: 0.
        input_order (str): Whether the input order is 'HWC' or 'CHW'.
            Default: 'CHW'.
        convert_to (str): Whether to convert the images to other color models.
            If None, the images are not altered. When computing for 'Y',
            the images are assumed to be in BGR order. Options are 'Y' and
            None. Default: None.

    Metrics:
        - SNR (float): Signal-to-Noise Ratio
    """

    metric = 'SNR'

    def __init__(self,
                 gt_key: str = 'gt_img',
                 pred_key: str = 'pred_img',
                 collect_device: str = 'cpu',
                 prefix: Optional[str] = None,
                 crop_border=0,
                 input_order='CHW',
                 convert_to=None) -> None:
        super().__init__(
            gt_key=gt_key,
            pred_key=pred_key,
            mask_key=None,
            collect_device=collect_device,
            prefix=prefix)

        self.crop_border = crop_border
        self.input_order = input_order
        self.convert_to = convert_to

    def process_image(self, gt, pred, mask):
        """Process an image.

        Args:
            gt (Torch | np.ndarray): GT image.
            pred (Torch | np.ndarray): Pred image.
            mask (Torch | np.ndarray): Mask of evaluation.
        Returns:
            np.ndarray: SNR result.
        """

        return snr(
            gt=gt,
            pred=pred,
            crop_border=self.crop_border,
            input_order=self.input_order,
            convert_to=self.convert_to,
            channel_order=self.channel_order)


def snr(gt,
        pred,
        crop_border=0,
        input_order='HWC',
        convert_to=None,
        channel_order='rgb'):
    """Calculate PSNR (Peak Signal-to-Noise Ratio).

    Ref: https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio

    Args:
        gt (ndarray): Images with range [0, 255].
        pred (ndarray): Images with range [0, 255].
        crop_border (int): Cropped pixels in each edges of an image. These
            pixels are not involved in the PSNR calculation. Default: 0.
        input_order (str): Whether the input order is 'HWC' or 'CHW'.
            Default: 'HWC'.
        convert_to (str): Whether to convert the images to other color models.
            If None, the images are not altered. When computing for 'Y',
            the images are assumed to be in BGR order. Options are 'Y' and
            None. Default: None.
        channel_order (str): The channel order of image. Default: 'rgb'.

    Returns:
        float: SNR result.
    """

    assert gt.shape == pred.shape, (
        f'Image shapes are different: {gt.shape}, {pred.shape}.')

    gt = img_transform(
        gt,
        crop_border=crop_border,
        input_order=input_order,
        convert_to=convert_to,
        channel_order=channel_order)
    pred = img_transform(
        pred,
        crop_border=crop_border,
        input_order=input_order,
        convert_to=convert_to,
        channel_order=channel_order)

    signal = ((gt)**2).mean()
    noise = ((gt - pred)**2).mean()

    result = 10. * np.log10(signal / noise)

    return result



# Copyright (c) OpenMMLab. All rights reserved.
import mmengine
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine.utils.dl_utils import TORCH_VERSION
from mmengine.utils.version_utils import digit_version


class ResnetBlock2D(nn.Module):
    """resnet block support down sample and up sample.

    Args:
        in_channels (int): input channels.
        out_channels (int): output channels.
        conv_shortcut (bool): whether to use conv shortcut.
        dropout (float): dropout rate.
        temb_channels (int): time embedding channels.
        groups (int): conv groups.
        groups_out (int): conv out groups.
        pre_norm (bool): whether to norm before conv. Todo: remove.
        eps (float): eps for groupnorm.
        non_linearity (str): non linearity type.
        time_embedding_norm (str): time embedding norm type.
        output_scale_factor (float): factor to scale input and output.
        use_in_shortcut (bool): whether to use conv in shortcut.
        up (bool): whether to upsample.
        down (bool): whether to downsample.
    """

    def __init__(
        self,
        in_channels,
        out_channels=None,
        conv_shortcut=False,
        dropout=0.0,
        temb_channels=512,
        groups=32,
        groups_out=None,
        pre_norm=True,
        eps=1e-6,
        non_linearity='silu',
        time_embedding_norm='default',
        output_scale_factor=1.0,
        use_in_shortcut=None,
        up=False,
        down=False,
    ):
        super().__init__()
        self.pre_norm = pre_norm
        self.pre_norm = True
        self.in_channels = in_channels
        out_channels = in_channels if out_channels is None else out_channels
        self.out_channels = out_channels
        self.use_conv_shortcut = conv_shortcut
        self.time_embedding_norm = time_embedding_norm
        self.up = up
        self.down = down
        self.output_scale_factor = output_scale_factor

        if groups_out is None:
            groups_out = groups

        self.norm1 = torch.nn.GroupNorm(
            num_groups=groups, num_channels=in_channels, eps=eps, affine=True)

        self.conv1 = torch.nn.Conv2d(
            in_channels, out_channels, kernel_size=3, stride=1, padding=1)

        if temb_channels is not None:
            self.time_emb_proj = torch.nn.Linear(temb_channels, out_channels)
        else:
            self.time_emb_proj = None

        self.norm2 = torch.nn.GroupNorm(
            num_groups=groups_out,
            num_channels=out_channels,
            eps=eps,
            affine=True)
        self.dropout = torch.nn.Dropout(dropout)
        self.conv2 = torch.nn.Conv2d(
            out_channels, out_channels, kernel_size=3, stride=1, padding=1)

        if non_linearity == 'silu' and \
                digit_version(TORCH_VERSION) > digit_version('1.6.0'):
            self.nonlinearity = nn.SiLU()
        else:
            mmengine.print_log('\'SiLU\' is not supported for '
                               f'torch < 1.6.0, found \'{torch.version}\'.'
                               'Use ReLu instead but result maybe wrong')
            self.nonlinearity = nn.ReLU()

        self.upsample = self.downsample = None
        if self.up:
            self.upsample = Upsample2D(in_channels, use_conv=False)
        elif self.down:
            self.downsample = Downsample2D(
                in_channels, use_conv=False, padding=1, name='op')

        self.use_in_shortcut = self.in_channels != self.out_channels if use_in_shortcut is None else use_in_shortcut  # noqa

        self.conv_shortcut = None
        if self.use_in_shortcut:
            self.conv_shortcut = torch.nn.Conv2d(
                in_channels, out_channels, kernel_size=1, stride=1, padding=0)

    def forward(self, input_tensor, temb):
        """forward with hidden states and time embeddings."""
        hidden_states = input_tensor

        hidden_states = self.norm1(hidden_states)
        hidden_states = self.nonlinearity(hidden_states)

        if self.upsample is not None:
            # upsample_nearest_nhwc fails with large batch sizes.
            # see https://github.com/huggingface/diffusers/issues/984
            if hidden_states.shape[0] >= 64:
                input_tensor = input_tensor.contiguous()
                hidden_states = hidden_states.contiguous()
            input_tensor = self.upsample(input_tensor)
            hidden_states = self.upsample(hidden_states)
        elif self.downsample is not None:
            input_tensor = self.downsample(input_tensor)
            hidden_states = self.downsample(hidden_states)

        hidden_states = self.conv1(hidden_states)

        if temb is not None:
            temb = self.time_emb_proj(self.nonlinearity(temb))[:, :, None,
                                                               None]
            hidden_states = hidden_states + temb

        hidden_states = self.norm2(hidden_states)
        hidden_states = self.nonlinearity(hidden_states)

        hidden_states = self.dropout(hidden_states)
        hidden_states = self.conv2(hidden_states)

        if self.conv_shortcut is not None:
            input_tensor = self.conv_shortcut(input_tensor)

        output_tensor = (input_tensor +
                         hidden_states) / self.output_scale_factor

        return output_tensor


class Upsample2D(nn.Module):
    """An upsampling layer with an optional convolution.

    Args:
        channels (int): channels in the inputs and outputs.
        use_conv (bool): a bool determining if a convolution is applied.
        use_conv_transpose (bool): whether to use conv transpose.
        out_channels (int): output channels.
    """

    def __init__(self,
                 channels,
                 use_conv=False,
                 use_conv_transpose=False,
                 out_channels=None,
                 name='conv'):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.use_conv_transpose = use_conv_transpose
        self.name = name

        conv = None
        if use_conv:
            conv = nn.Conv2d(self.channels, self.out_channels, 3, padding=1)
        else:
            conv = nn.ConvTranspose2d(channels, self.out_channels, 4, 2, 1)

        self.conv = conv

    def forward(self, hidden_states, output_size=None):
        """forward with hidden states."""
        assert hidden_states.shape[1] == self.channels

        if self.use_conv_transpose:
            return self.conv(hidden_states)

        # if `output_size` is passed we force the interpolation output
        # size and do not make use of `scale_factor=2`
        if output_size is None:
            hidden_states = F.interpolate(
                hidden_states, scale_factor=2.0, mode='nearest')
        else:
            hidden_states = F.interpolate(
                hidden_states, size=output_size, mode='nearest')

        hidden_states = self.conv(hidden_states)

        return hidden_states


class Downsample2D(nn.Module):
    """A downsampling layer with an optional convolution.

    Args:
        channels (int): channels in the inputs and outputs.
        use_conv (bool): a bool determining if a convolution is applied.
        out_channels (int): output channels
        padding (int): padding num
    """

    def __init__(self,
                 channels,
                 use_conv=False,
                 out_channels=None,
                 padding=1,
                 name='conv'):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.padding = padding
        stride = 2
        self.name = name

        if use_conv:
            conv = nn.Conv2d(
                self.channels,
                self.out_channels,
                3,
                stride=stride,
                padding=padding)
        else:
            assert self.channels == self.out_channels
            conv = nn.AvgPool2d(kernel_size=stride, stride=stride)

        self.conv = conv

    def forward(self, hidden_states):
        """forward with hidden states."""
        assert hidden_states.shape[1] == self.channels
        if self.use_conv and self.padding == 0:
            pad = (0, 1, 0, 1)
            hidden_states = F.pad(hidden_states, pad, mode='constant', value=0)

        assert hidden_states.shape[1] == self.channels
        hidden_states = self.conv(hidden_states)

        return hidden_states



# Copyright (c) OpenMMLab. All rights reserved.
import mmcv
import mmengine

from .version import __version__, version_info

try:
    from mmcv.utils import digit_version
except ImportError:

    def digit_version(version_str):
        digit_ver = []
        for x in version_str.split('.'):
            if x.isdigit():
                digit_ver.append(int(x))
            elif x.find('rc') != -1:
                patch_version = x.split('rc')
                digit_ver.append(int(patch_version[0]) - 1)
                digit_ver.append(int(patch_version[1]))
        return digit_ver


MMCV_MIN = '2.0.0rc1'
MMCV_MAX = '2.1.0'
mmcv_min_version = digit_version(MMCV_MIN)
mmcv_max_version = digit_version(MMCV_MAX)
mmcv_version = digit_version(mmcv.__version__)

MMENGINE_MIN = '0.4.0'
MMENGINE_MAX = '1.0.0'
mmengine_min_version = digit_version(MMENGINE_MIN)
mmengine_max_version = digit_version(MMENGINE_MAX)
mmengine_version = digit_version(mmengine.__version__)

assert (mmcv_min_version <= mmcv_version < mmcv_max_version), \
    f'mmcv=={mmcv.__version__} is used but incompatible. ' \
    f'Please install mmcv-full>={mmcv_min_version}, <{mmcv_max_version}.'

assert (mmengine_min_version <= mmengine_version < mmengine_max_version), \
    f'mmengine=={mmengine.__version__} is used but incompatible. ' \
    f'Please install mmengine>={mmengine_min_version}, ' \
    f'<{mmengine_max_version}.'

__all__ = ['__version__', 'version_info']



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Union

import numpy as np
import torch

from mmagic.models.utils.diffusion_utils import betas_for_alpha_bar
from mmagic.registry import DIFFUSION_SCHEDULERS


@DIFFUSION_SCHEDULERS.register_module()
class EditDDIMScheduler:
    """```EditDDIMScheduler``` support the diffusion and reverse process
    formulated in https://arxiv.org/abs/2010.02502.

    The code is heavily influenced by https://github.com/huggingface/diffusers/blob/main/src/diffusers/schedulers/scheduling_ddim.py. # noqa
    The difference is that we ensemble gradient-guided sampling in step function.

    Args:
        num_train_timesteps (int, optional): _description_. Defaults to 1000.
        beta_start (float, optional): _description_. Defaults to 0.0001.
        beta_end (float, optional): _description_. Defaults to 0.02.
        beta_schedule (str, optional): _description_. Defaults to "linear".
        variance_type (str, optional): _description_. Defaults to 'learned_range'.
        timestep_values (_type_, optional): _description_. Defaults to None.
        clip_sample (bool, optional): _description_. Defaults to True.
        set_alpha_to_one (bool, optional): _description_. Defaults to True.
    """

    def __init__(
        self,
        num_train_timesteps=1000,
        beta_start=0.0001,
        beta_end=0.02,
        beta_schedule='linear',
        variance_type='learned_range',
        timestep_values=None,
        clip_sample=True,
        set_alpha_to_one=True,
    ):
        self.num_train_timesteps = num_train_timesteps
        self.beta_start = beta_start
        self.beta_end = beta_end
        self.beta_schedule = beta_schedule
        self.variance_type = variance_type
        self.timestep_values = timestep_values
        self.clip_sample = clip_sample
        self.set_alpha_to_one = set_alpha_to_one

        if beta_schedule == 'linear':
            self.betas = np.linspace(
                beta_start, beta_end, num_train_timesteps, dtype=np.float32)
        elif beta_schedule == 'scaled_linear':
            # this schedule is very specific to the latent diffusion model.
            self.betas = np.linspace(
                beta_start**0.5,
                beta_end**0.5,
                num_train_timesteps,
                dtype=np.float32)**2
        elif beta_schedule == 'squaredcos_cap_v2':
            # Glide cosine schedule
            self.betas = betas_for_alpha_bar(num_train_timesteps)
        else:
            raise NotImplementedError(
                f'{beta_schedule} does is not implemented for {self.__class__}'
            )

        self.alphas = 1.0 - self.betas
        self.alphas_cumprod = np.cumprod(self.alphas, axis=0)

        # At every step in ddim, we are looking into the
        # previous alphas_cumprod. For the final step,
        # there is no previous alphas_cumprod because we are already
        # at 0 `set_alpha_to_one` decides whether we set this parameter
        # simply to one or whether we use the final alpha of the
        # "non-previous" one.
        self.final_alpha_cumprod = np.array(
            1.0) if set_alpha_to_one else self.alphas_cumprod[0]

        # standard deviation of the initial noise distribution
        self.init_noise_sigma = 1.0

        # setable values
        self.num_inference_steps = None
        self.timesteps = np.arange(0, num_train_timesteps)[::-1].copy()

    def set_timesteps(self, num_inference_steps, offset=0):
        """set time steps."""

        self.num_inference_steps = num_inference_steps
        self.timesteps = np.arange(
            0, self.num_train_timesteps,
            self.num_train_timesteps // self.num_inference_steps)[::-1].copy()
        self.timesteps += offset

    def scale_model_input(self,
                          sample: torch.FloatTensor,
                          timestep: Optional[int] = None) -> torch.FloatTensor:
        """Ensures interchangeability with schedulers that need to scale the
        denoising model input depending on the current timestep.

        Args:
            sample (`torch.FloatTensor`): input sample
            timestep (`int`, optional): current timestep

        Returns:
            `torch.FloatTensor`: scaled input sample
        """

        return sample

    def _get_variance(self, timestep, prev_timestep):
        """get variance."""

        alpha_prod_t = self.alphas_cumprod[timestep]
        alpha_prod_t_prev = self.alphas_cumprod[
            prev_timestep] if prev_timestep >= 0 else self.final_alpha_cumprod
        beta_prod_t = 1 - alpha_prod_t
        beta_prod_t_prev = 1 - alpha_prod_t_prev
        variance = (beta_prod_t_prev /
                    beta_prod_t) * (1 - alpha_prod_t / alpha_prod_t_prev)
        return variance

    def step(
        self,
        model_output: Union[torch.FloatTensor, np.ndarray],
        timestep: int,
        sample: Union[torch.FloatTensor, np.ndarray],
        cond_fn=None,
        cond_kwargs={},
        eta: float = 0.0,
        use_clipped_model_output: bool = False,
        generator=None,
    ):
        """step forward."""

        output = {}
        if self.num_inference_steps is None:
            raise ValueError("Number of inference steps is 'None', '\
                    'you need to run 'set_timesteps' '\
                        'after creating the scheduler")

        pred = None
        if isinstance(model_output, dict):
            pred = model_output['pred']
            model_output = model_output['eps']
        elif model_output.shape[1] == sample.shape[
                1] * 2 and self.variance_type in ['learned', 'learned_range']:
            model_output, _ = torch.split(model_output, sample.shape[1], dim=1)
        else:
            if not model_output.shape == sample.shape:
                raise TypeError

        # See formulas (12) and (16) of DDIM paper https://arxiv.org/pdf/2010.02502.pdf # noqa
        # Ideally, read DDIM paper in-detail understanding

        # Notation (<variable name> -> <name in paper>
        # - pred_noise_t -> e_theta(x_t, t)
        # - pred_original_sample -> f_theta(x_t, t) or x_0
        # - std_dev_t -> sigma_t
        # - eta -> η
        # - pred_sample_direction -> "direction pointingc to x_t"
        # - pred_prev_sample -> "x_t-1"

        # 1. get previous step value (=t-1)
        prev_timestep = (
            timestep - self.num_train_timesteps // self.num_inference_steps)

        # 2. compute alphas, betas
        alpha_prod_t = self.alphas_cumprod[timestep]
        alpha_prod_t_prev = self.alphas_cumprod[
            prev_timestep] if prev_timestep >= 0 else self.final_alpha_cumprod
        beta_prod_t = 1 - alpha_prod_t

        # 3. compute predicted original sample from predicted noise also called
        # "predicted x_0" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf # noqa
        pred_original_sample = (sample - (
            (beta_prod_t)**(0.5)) * model_output) / alpha_prod_t**(0.5)
        if pred is not None:
            pred_original_sample = pred

        gradient = 0.
        if cond_fn is not None:
            if cond_fn.__name__ == 'classifier_grad':
                y = cond_kwargs['y']
                classifier = cond_kwargs['classifier']
                classifier_scale = cond_kwargs['classifier_scale']
                gradient = cond_fn(
                    classifier,
                    sample,
                    timestep,
                    y=y,
                    classifier_scale=classifier_scale)
            else:
                gradient = cond_fn(
                    cond_kwargs.pop('unet'), self, sample, timestep,
                    beta_prod_t, cond_kwargs.pop('model_stats'), **cond_kwargs)
            model_output = model_output - (beta_prod_t**0.5) * gradient
            pred_original_sample = (
                sample -
                (beta_prod_t**(0.5)) * model_output) / alpha_prod_t**(0.5)
        # 4. Clip "predicted x_0"
        if self.clip_sample:
            pred_original_sample = torch.clamp(pred_original_sample, -1, 1)

        # 5. compute variance: "sigma_t(η)" -> see formula (16)
        # σ_t = sqrt((1 − α_t−1)/(1 − α_t)) * sqrt(1 − α_t/α_t−1)
        variance = self._get_variance(timestep, prev_timestep)
        std_dev_t = eta * variance**(0.5)
        output.update(dict(sigma=std_dev_t))

        if use_clipped_model_output:
            # the model_output is always
            # re-derived from the clipped x_0 in Glide
            model_output = (sample - (alpha_prod_t**(0.5)) *
                            pred_original_sample) / beta_prod_t**(0.5)

        # 6. compute "direction pointing to x_t" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf # noqa
        pred_sample_direction = (1 - alpha_prod_t_prev -
                                 std_dev_t**2)**(0.5) * model_output

        # 7. compute x_t without "random noise" of
        # formula (12) from https://arxiv.org/pdf/2010.02502.pdf
        prev_mean = alpha_prod_t_prev**(
            0.5) * pred_original_sample + pred_sample_direction
        output.update(dict(mean=prev_mean, prev_sample=prev_mean))

        if eta > 0:
            device = model_output.device if torch.is_tensor(
                model_output) else 'cpu'
            noise = torch.randn(
                model_output.shape, generator=generator).to(device)
            variance = std_dev_t * noise

            if not torch.is_tensor(model_output):
                variance = variance.numpy()

            prev_sample = prev_mean + variance
            output.update({'prev_sample': prev_sample})

        # NOTE: this x0 is twice computed
        output.update({
            'original_sample': pred_original_sample,
            'beta_prod_t': beta_prod_t
        })
        return output

    def add_noise(self, original_samples, noise, timesteps):
        """add noise."""

        sqrt_alpha_prod = self.alphas_cumprod[timesteps]**0.5
        sqrt_one_minus_alpha_prod = (1 - self.alphas_cumprod[timesteps])**0.5
        noisy_samples = (
            sqrt_alpha_prod * original_samples +
            sqrt_one_minus_alpha_prod * noise)
        return noisy_samples

    def __len__(self):
        return self.num_train_timesteps



sigma = 15
test_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='RandomNoise',
        params=dict(
            noise_type=['gaussian'],
            noise_prob=[1],
            gaussian_sigma=[sigma, sigma],
            gaussian_gray_noise_prob=0),
        keys=['img']),
    dict(type='PackInputs')
]

data_root = 'data/denoising_gaussian_test'
cbsd68_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='CBSD68', task_name='denoising'),
        data_root=data_root,
        data_prefix=dict(img='CBSD68', gt='CBSD68'),
        pipeline=test_pipeline))
cbsd68_evaluator = [
    dict(type='PSNR', prefix='CBSD68'),
    dict(type='SSIM', prefix='CBSD68'),
]

kodak24_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='Kodak24', task_name='denoising'),
        data_root=data_root,
        data_prefix=dict(img='Kodak24', gt='Kodak24'),
        pipeline=test_pipeline))
kodak24_evaluator = [
    dict(type='PSNR', prefix='Kodak24'),
    dict(type='SSIM', prefix='Kodak24'),
]

mcmaster_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='McMaster', task_name='denoising'),
        data_root=data_root,
        data_prefix=dict(img='McMaster', gt='McMaster'),
        pipeline=test_pipeline))
mcmaster_evaluator = [
    dict(type='PSNR', prefix='McMaster'),
    dict(type='SSIM', prefix='McMaster'),
]

urban100_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='Urban100', task_name='denoising'),
        data_root=data_root,
        data_prefix=dict(img='Urban100', gt='Urban100'),
        pipeline=test_pipeline))
urban100_evaluator = [
    dict(type='PSNR', prefix='Urban100'),
    dict(type='SSIM', prefix='Urban100'),
]

# test config
test_cfg = dict(type='MultiTestLoop')
test_dataloader = [
    cbsd68_dataloader,
    kodak24_dataloader,
    mcmaster_dataloader,
    urban100_dataloader,
]
test_evaluator = [
    cbsd68_evaluator,
    kodak24_evaluator,
    mcmaster_evaluator,
    urban100_evaluator,
]



# Copyright (c) OpenMMLab. All rights reserved.
import cv2
import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets
from utils import layout2widget


class SliderTab(QtWidgets.QWidget):

    def __init__(self):
        super().__init__()
        self.images = [None, None]

        self.cb_1 = QtWidgets.QComboBox()
        self.cb_2 = QtWidgets.QComboBox()
        self.cb_1.currentIndexChanged.connect(self.change_image)
        self.cb_2.currentIndexChanged.connect(self.change_image)
        self.btn_add = QtWidgets.QPushButton()
        self.btn_add.setText('Add')
        self.btn_add.clicked.connect(self.add)
        left_grid = QtWidgets.QGridLayout()
        left_grid.addWidget(self.cb_1, 0, 0)
        left_grid.addWidget(self.cb_2, 1, 0)
        left_grid.addWidget(self.btn_add, 2, 0)

        self.imageArea = QtWidgets.QLabel()
        self.imageArea.setFrameShape(QtWidgets.QFrame.Box)
        self.imageArea.setLineWidth(2)
        self.imageArea.setAlignment(QtCore.Qt.AlignBottom)
        self.imageArea.setStyleSheet(
            'border-width: 0px; border-style: solid; border-color: rgb(100, 100, 100);background-color: rgb(255, 255, 255)'  # noqa
        )
        self.slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider.setMaximum(800)
        self.slider.valueChanged.connect(self.show_image)
        right_grid = QtWidgets.QGridLayout()
        right_grid.addWidget(self.imageArea, 0, 0)
        right_grid.addWidget(self.slider, 1, 0)

        # Splitter
        hsplitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        hsplitter.addWidget(layout2widget(left_grid))
        hsplitter.addWidget(layout2widget(right_grid))
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(hsplitter)
        self.setLayout(hlayout)

    def add(self):
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Select gt file', '', 'Images (*.jpg *.png *.mp4 *.avi)')
        if self.cb_1.count() > 0:
            if self.cb_1.findText(path) > -1:
                self.cb_1.removeItem(self.cb_1.findText(path))
            if self.cb_2.findText(path) > -1:
                self.cb_2.removeItem(self.cb_2.findText(path))
            self.cb_1.addItem(path)
            self.cb_2.addItem(path)
            self.cb_2.setCurrentIndex(self.cb_2.count() - 1)
        else:
            self.cb_1.addItem(path)
            self.cb_2.addItem(path)
            self.cb_1.setCurrentIndex(0)
            self.cb_2.setCurrentIndex(0)

    def change_image(self):
        self.images[0] = cv2.imread(self.cb_1.currentText())
        self.images[1] = cv2.imread(self.cb_2.currentText())
        if self.images[0] is None or self.images[1] is None:
            return
        self.show_image()

    def show_image(self):
        img1, img2 = self.images
        h2, w2, c2 = img2.shape
        img2 = cv2.resize(img2, (800, int(800 / w2 * h2)))
        h2, w2, c2 = img2.shape
        img1 = cv2.resize(img1, (w2, h2))
        v = self.slider.value()
        img11 = img1[:, 0:v].copy()
        img22 = img2[:, v:].copy()
        img = np.hstack((img11, img22))
        img = cv2.line(img, (v, 0), (v, h2), (0, 222, 0), 4)
        rgb_img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img_dis = QtGui.QImage(rgb_img, w2, h2, w2 * c2,
                               QtGui.QImage.Format_RGB888)
        jpg = QtGui.QPixmap.fromImage(img_dis).scaled(
            self.imageArea.width(), int(self.imageArea.width() / w2 * h2))
        self.imageArea.setPixmap(jpg)


class GeneralPage(QtWidgets.QWidget):

    def __init__(self) -> None:
        super().__init__()

        self.tab_slider = SliderTab()
        self.tabs = QtWidgets.QTabWidget()
        self.tabs.addTab(self.tab_slider, 'before/after slider')

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.tabs)



# Copyright (c) OpenMMLab. All rights reserved.
import hashlib
import os
import os.path as osp
import pickle
import sys
from contextlib import contextmanager
from copy import deepcopy
from typing import Optional, Tuple

import mmengine
import numpy as np
import torch
import torch.nn as nn
from mmengine import is_filepath, print_log
from mmengine.dataset import BaseDataset, Compose, pseudo_collate
from mmengine.dist import (all_gather, get_dist_info, get_world_size,
                           is_main_process)
from mmengine.evaluator import BaseMetric
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset
from torchvision.models.inception import inception_v3

from mmagic.utils import MMAGIC_CACHE_DIR, download_from_url
from . import InceptionV3

ALLOWED_INCEPTION = ['StyleGAN', 'PyTorch']
TERO_INCEPTION_URL = 'https://nvlabs-fi-cdn.nvidia.com/stylegan2-ada-pytorch/pretrained/metrics/inception-2015-12-05.pt'  # noqa


@contextmanager
def disable_gpu_fuser_on_pt19():
    """On PyTorch 1.9 a CUDA fuser bug prevents the Inception JIT model to run.

    Refers to:
      https://github.com/GaParmar/clean-fid/blob/5e1e84cdea9654b9ac7189306dfa4057ea2213d8/cleanfid/inception_torchscript.py#L9  # noqa
      https://github.com/GaParmar/clean-fid/issues/5
      https://github.com/pytorch/pytorch/issues/64062
    """
    if torch.__version__.startswith('1.9.'):
        old_val = torch._C._jit_can_fuse_on_gpu()
        torch._C._jit_override_can_fuse_on_gpu(False)
    yield
    if torch.__version__.startswith('1.9.'):
        torch._C._jit_override_can_fuse_on_gpu(old_val)


def load_inception(inception_args, metric):
    """Load Inception Model from given ``inception_args`` and ``metric``.

    This function would try to load Inception under the guidance of 'type'
    given in `inception_args`, if not given, we would try best to load Tero's
    ones. In detail, we would first try to load the model from disk with
    the given 'inception_path', and then try to download the checkpoint from
    'inception_url'. If both method are failed, pytorch version of Inception
    would be loaded.

    Args:
        inception_args (dict): Keyword args for inception net.
        metric (string): Metric to use the Inception. This argument would
            influence the pytorch's Inception loading.

    Returns:
        model (torch.nn.Module): Loaded Inception model.
        style (string): The version of the loaded Inception.
    """

    if not isinstance(inception_args, dict):
        raise TypeError('Receive invalid \'inception_args\': '
                        f'\'{inception_args}\'')

    _inception_args = deepcopy(inception_args)
    inception_type = _inception_args.pop('type', None)

    if torch.__version__ < '1.6.0':
        print_log(
            'Current Pytorch Version not support script module, load '
            'Inception Model from torch model zoo. If you want to use '
            'Tero\' script model, please update your Pytorch higher '
            f'than \'1.6\' (now is {torch.__version__})', 'current')
        return _load_inception_torch(_inception_args, metric), 'pytorch'

    # load pytorch version is specific
    if inception_type != 'StyleGAN':
        return _load_inception_torch(_inception_args, metric), 'pytorch'

    # try to load Tero's version
    path = _inception_args.get('inception_path', TERO_INCEPTION_URL)
    if path is None:
        path = TERO_INCEPTION_URL

    # try to parse `path` as web url and download
    if 'http' not in path:
        model = _load_inception_from_path(path)
        if isinstance(model, torch.nn.Module):
            return model, 'StyleGAN'

    # try to parse `path` as path on disk
    model = _load_inception_from_url(path)
    if isinstance(model, torch.nn.Module):
        return model, 'StyleGAN'

    raise RuntimeError('Cannot Load Inception Model, please check the input '
                       f'`inception_args`: {inception_args}')


def _load_inception_from_path(inception_path):
    """Load inception from passed path.

    Args:
        inception_path (str): The path of inception.

    Returns:
        nn.Module: The loaded inception.
    """
    print_log(
        'Try to load Tero\'s Inception Model from '
        f'\'{inception_path}\'.', 'current')
    try:
        model = torch.jit.load(inception_path)
        print_log('Load Tero\'s Inception Model successfully.', 'current')
    except Exception as e:
        model = None
        print_log('Load Tero\'s Inception Model failed. '
                  f'\'{e}\' occurs.', 'current')
    return model


def _load_inception_from_url(inception_url: str) -> nn.Module:
    """Load Inception network from the give `inception_url`"""
    inception_url = inception_url if inception_url else TERO_INCEPTION_URL
    print_log(f'Try to download Inception Model from {inception_url}...',
              'current')
    try:
        path = download_from_url(inception_url, dest_dir=MMAGIC_CACHE_DIR)
        print_log('Download Finished.', 'current')
        return _load_inception_from_path(path)
    except Exception as e:
        print_log(f'Download Failed. {e} occurs.', 'current')
        return None


def _load_inception_torch(inception_args, metric) -> nn.Module:
    """Load Inception network from PyTorch's model zoo."""
    assert metric in ['FID', 'IS']
    if metric == 'FID':
        inception_model = InceptionV3([3], **inception_args)
    elif metric == 'IS':
        inception_model = inception_v3(pretrained=True, transform_input=False)
        print_log(
            'Load Inception V3 Network from Pytorch Model Zoo '
            'for IS calculation. The results can only used '
            'for monitoring purposes. To get more accuracy IS, '
            'please use Tero\'s Inception V3 checkpoints '
            'and use Bicubic Interpolation with Pillow backend '
            'for image resizing. More details may refer to '
            'https://github.com/open-mmlab/MMEditing/blob/master/docs/en/quick_run.md#is.',  # noqa
            'current')
    return inception_model


def get_inception_feat_cache_name_and_args(dataloader: DataLoader,
                                           metric: BaseMetric, real_nums: int,
                                           capture_mean_cov: bool,
                                           capture_all: bool
                                           ) -> Tuple[str, dict]:
    """Get the name and meta info of the inception feature cache file
    corresponding to the input dataloader and metric.

    The meta info includes
    'data_root', 'data_prefix', 'meta_info' and 'pipeline' of the dataset, and
    'inception_style' and 'inception_args' of the metric. Then we calculate the
    hash value of the meta info dict with md5, and the name of the inception
    feature cache will be 'inception_feat_{HASH}.pkl'.
    Args:
        dataloader (Dataloader): The dataloader of real images.
        metric (BaseMetric): The metric which needs inception features.
        real_nums (int): Number of images used to extract inception feature.
        capture_mean_cov (bool): Whether save the mean and covariance of
            inception feature. Defaults to False.
        capture_all (bool): Whether save the raw inception feature. Defaults
            to False.
    Returns:
        Tuple[str, dict]: Filename and meta info dict of the inception feature
            cache.
    """

    dataset: BaseDataset = dataloader.dataset
    assert isinstance(dataset, Dataset), (
        f'Only support normal dataset, but receive {type(dataset)}.')

    # get dataset info
    data_root = deepcopy(dataset.data_root)
    data_prefix = deepcopy(dataset.data_prefix)
    metainfo = dataset.metainfo
    pipeline = repr(dataset.pipeline)

    # get metric info
    inception_style = metric.inception_style
    inception_args = getattr(metric, 'inception_args', None)

    real_key = 'gt_img' if metric.real_key is None else metric.real_key
    args = dict(
        data_root=data_root,
        data_prefix=data_prefix,
        metainfo=metainfo,
        pipeline=pipeline,
        inception_style=inception_style,
        inception_args=inception_args,
        # save `num_gpus` because this may influence the data loading order
        num_gpus=get_world_size(),
        capture_mean_cov=capture_mean_cov,
        capture_all=capture_all,
        real_keys=real_key,
        real_nums=real_nums)

    real_nums_str = 'full' if real_nums == -1 else str(real_nums)
    md5 = hashlib.md5(repr(sorted(args.items())).encode('utf-8'))
    if capture_all:
        prefix = 'inception_state-capture_all'
    elif capture_mean_cov:
        prefix = 'inception_state-capture_mean_cov'
    else:
        prefix = 'inception_state-capture_all_mean_cov'
    cache_tag = f'{prefix}-{real_nums_str}-{md5.hexdigest()}.pkl'
    return cache_tag, args


def get_vgg_feat_cache_name_and_args(dataloader: DataLoader,
                                     metric: BaseMetric) -> Tuple[str, dict]:
    """Get the name and meta info of the vgg feature cache file corresponding
    to the input dataloader and metric.

    The meta info includes 'data_root',
    'data_prefix', 'meta_info' and 'pipeline' of the dataset, and
    'use_tero_scirpt' of the metric. Then we calculate the hash value of the
    meta info dict with md5, and the name of the vgg feature cache will be
    'vgg_feat_{HASH}.pkl'.
    Args:
        dataloader (Dataloader): The dataloader of real images.
        metric (BaseMetric): The metric which needs inception features.
    Returns:
        Tuple[str, dict]: Filename and meta info dict of the inception feature
            cache.
    """

    dataset: BaseDataset = dataloader.dataset
    assert isinstance(dataset, Dataset), (
        f'Only support normal dataset, but receive {type(dataset)}.')

    # get dataset info
    data_root = deepcopy(dataset.data_root)
    data_prefix = deepcopy(dataset.data_prefix)
    metainfo = dataset.metainfo
    pipeline = dataset.pipeline
    if isinstance(pipeline, Compose):
        pipeline_str = repr(pipeline)
    else:
        pipeline_str = ''

    # get metric info
    use_tero_scirpt = metric.use_tero_scirpt

    args = dict(
        data_root=data_root,
        data_prefix=data_prefix,
        metainfo=metainfo,
        pipeline=pipeline_str,
        use_tero_scirpt=use_tero_scirpt)

    md5 = hashlib.md5(repr(sorted(args.items())).encode('utf-8'))
    cache_tag = f'vgg_state-{md5.hexdigest()}.pkl'
    return cache_tag, args


def prepare_inception_feat(dataloader: DataLoader,
                           metric: BaseMetric,
                           data_preprocessor: Optional[nn.Module] = None,
                           capture_mean_cov: bool = False,
                           capture_all: bool = False) -> dict:
    """Prepare inception feature for the input metric.

    - If `metric.inception_pkl` is an online path, try to download and load
      it. If cannot download or load, corresponding error will be raised.
    - If `metric.inception_pkl` is local path and file exists, try to load the
      file. If cannot load, corresponding error will be raised.
    - If `metric.inception_pkl` is local path and file not exists, we will
      extract the inception feature manually and save to 'inception_pkl'.
    - If `metric.inception_pkl` is not defined, we will extract the inception
      feature and save it to default cache dir with default name.

    Args:
        dataloader (Dataloader): The dataloader of real images.
        metric (BaseMetric): The metric which needs inception features.
        data_preprocessor (Optional[nn.Module]): Data preprocessor of the
            module. Used to preprocess the real images. If not passed, real
            images will automatically normalized to [-1, 1]. Defaults to None.
        capture_mean_cov (bool): Whether save the mean and covariance of
            inception feature. Defaults to False.
        capture_all (bool): Whether save the raw inception feature. If true,
            it will take a lot of time to save the inception feature. Defaults
            to False.

    Returns:
        dict: Dict contains inception feature.
    """
    assert capture_mean_cov or capture_all, (
        'At least one of \'capture_mean_cov\' and \'capture_all\' is True.')
    if not hasattr(metric, 'inception_pkl'):
        return
    inception_pkl: Optional[str] = metric.inception_pkl

    if isinstance(inception_pkl, str):
        if is_filepath(inception_pkl) and osp.exists(inception_pkl):
            with open(inception_pkl, 'rb') as file:
                inception_state = pickle.load(file)
            print_log(
                f'\'{metric.prefix}\' successful load inception feature '
                f'from \'{inception_pkl}\'', 'current')
            return inception_state
        elif inception_pkl.startswith('s3'):
            try:
                raise NotImplementedError(
                    'Not support download from Ceph currently')
            except Exception as exp:
                raise exp('Not support download from Ceph currently')
        elif inception_pkl.startswith('http'):
            try:
                raise NotImplementedError(
                    'Not support download from url currently')
            except Exception as exp:
                # cannot download, raise error
                raise exp('Not support download from url currently')

    # cannot load or download from file, extract manually
    assert hasattr(metric, 'real_nums'), (
        f'Metric \'{metric.name}\' must have attribute \'real_nums\'.')
    real_nums = metric.real_nums
    if inception_pkl is None:
        inception_pkl, args = get_inception_feat_cache_name_and_args(
            dataloader, metric, real_nums, capture_mean_cov, capture_all)
        inception_pkl = osp.join(MMAGIC_CACHE_DIR, inception_pkl)
    else:
        args = dict()
    if osp.exists(inception_pkl):
        with open(inception_pkl, 'rb') as file:
            real_feat = pickle.load(file)
        print_log(f'load preprocessed feat from {inception_pkl}', 'current')
        return real_feat

    assert hasattr(metric, 'inception'), (
        'Metric must have a inception network to extract inception features.')

    real_feat = []

    print_log(
        f'Inception pkl \'{inception_pkl}\' is not found, extract '
        'manually.', 'current')

    import rich.progress

    dataset, batch_size = dataloader.dataset, dataloader.batch_size
    if real_nums == -1:
        num_items = len(dataset)
    else:
        num_items = min(len(dataset), real_nums)

    rank, num_gpus = get_dist_info()
    item_subset = [(i * num_gpus + rank) % num_items
                   for i in range((num_items - 1) // num_gpus + 1)]
    inception_dataloader = DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=item_subset,
        collate_fn=pseudo_collate,
        shuffle=False,
        drop_last=False)
    # init rich pbar for the main process
    if is_main_process():
        # check the launcher
        slurm_env_name = ['SLURM_PROCID', 'SLURM_NTASKS', 'SLURM_NODELIST']
        if all([n in os.environ for n in slurm_env_name]):
            is_slurm = True
            pbar = mmengine.ProgressBar(len(inception_dataloader))
        else:
            is_slurm = False
            columns = [
                rich.progress.TextColumn('[bold blue]{task.description}'),
                rich.progress.BarColumn(bar_width=40),
                rich.progress.TaskProgressColumn(),
                rich.progress.TimeRemainingColumn(),
            ]
            pbar = rich.progress.Progress(*columns)
            pbar.start()
            task = pbar.add_task(
                'Calculate Inception Feature.',
                total=len(inception_dataloader),
                visible=True)

    for data in inception_dataloader:
        # set training = False to avoid norm + convert to BGR
        data_samples = data_preprocessor(data, False)['data_samples']

        real_key = 'gt_img' if metric.real_key is None else metric.real_key
        img = getattr(data_samples, real_key)

        real_feat_ = metric.forward_inception(img).cpu()
        real_feat.append(real_feat_)

        if is_main_process():
            if is_slurm:
                pbar.update(1)
            else:
                pbar.update(task, advance=1)

    # stop the pbar
    if is_main_process():
        if is_slurm:
            sys.stdout.write('\n')
        else:
            pbar.stop()

    # collect results
    real_feat = torch.cat(real_feat)
    # use `all_gather` here, gather tensor is much quicker than gather object.
    real_feat = all_gather(real_feat)

    # only cat on the main process
    if is_main_process():
        inception_state = dict(**args)
        if capture_mean_cov:
            real_feat = torch.cat(real_feat, dim=0)[:num_items].numpy()
            real_mean = np.mean(real_feat, 0)
            real_cov = np.cov(real_feat, rowvar=False)
            inception_state['real_mean'] = real_mean
            inception_state['real_cov'] = real_cov
        if capture_all:
            inception_state['raw_feature'] = real_feat
        dir_name = osp.dirname(inception_pkl)
        os.makedirs(dir_name, exist_ok=True)
        print_log(
            f'Saving inception pkl to {inception_pkl}. Please be patient.',
            'current')
        with open(inception_pkl, 'wb') as file:
            pickle.dump(inception_state, file)
        print_log('Inception pkl Finished.', 'current')
        return inception_state


def prepare_vgg_feat(dataloader: DataLoader,
                     metric: BaseMetric,
                     data_preprocessor: Optional[nn.Module] = None,
                     auto_save=True) -> np.ndarray:
    """Prepare vgg feature for the input metric.

    - If `metric.vgg_pkl` is an online path, try to download and load
      it. If cannot download or load, corresponding error will be raised.
    - If `metric.vgg_pkl` is local path and file exists, try to load the
      file. If cannot load, corresponding error will be raised.
    - If `metric.vgg_pkl` is local path and file not exists, we will
      extract the vgg feature manually and save to 'vgg_pkl'.
    - If `metric.vgg_pkl` is not defined, we will extract the vgg
      feature and save it to default cache dir with default name.

    Args:
        dataloader (Dataloader): The dataloader of real images.
        metric (BaseMetric): The metric which needs vgg features.
        data_preprocessor (Optional[nn.Module]): Data preprocessor of the
            module. Used to preprocess the real images. If not passed, real
            images will automatically normalized to [-1, 1]. Defaults to None.
        Returns:
            np.ndarray: Loaded vgg feature.
    """
    if not hasattr(metric, 'vgg16_pkl'):
        return
    vgg_pkl: Optional[str] = metric.vgg16_pkl

    if isinstance(vgg_pkl, str):
        if is_filepath(vgg_pkl) and osp.exists(vgg_pkl):
            with open(vgg_pkl, 'rb') as file:
                vgg_state = pickle.load(file)
            print_log(
                f'\'{metric.prefix}\' successful load VGG feature '
                f'from \'{vgg_pkl}\'', 'current')
            return vgg_state['vgg_feat']
        elif vgg_pkl.startswith('s3'):
            try:
                raise NotImplementedError(
                    'Not support download from Ceph currently')
            except Exception as exp:
                raise exp('Not support download from Ceph currently')
        elif vgg_pkl.startswith('http'):
            try:
                raise NotImplementedError(
                    'Not support download from url currently')
            except Exception as exp:
                # cannot download, raise error
                raise exp('Not support download from url currently')

    # cannot load or download from file, extract manually
    if vgg_pkl is None:
        vgg_pkl, args = get_vgg_feat_cache_name_and_args(dataloader, metric)
        vgg_pkl = osp.join(MMAGIC_CACHE_DIR, vgg_pkl)
    else:
        args = dict()
    if osp.exists(vgg_pkl):
        with open(vgg_pkl, 'rb') as file:
            real_feat = pickle.load(file)['vgg_feat']
        print(f'load preprocessed feat from {vgg_pkl}')
        return real_feat

    assert hasattr(
        metric,
        'vgg16'), ('Metric must have a vgg16 network to extract vgg features.')

    real_feat = []

    print_log(f'Vgg pkl \'{vgg_pkl}\' is not found, extract '
              'manually.', 'current')

    import rich.progress

    # init rich pbar for the main process
    if is_main_process():
        columns = [
            rich.progress.TextColumn('[bold blue]{task.description}'),
            rich.progress.BarColumn(bar_width=40),
            rich.progress.TaskProgressColumn(),
            rich.progress.TimeRemainingColumn(),
        ]
        pbar = rich.progress.Progress(*columns)
        pbar.start()
        task = pbar.add_task(
            'Calculate VGG16 Feature.',
            total=len(dataloader.dataset),
            visible=True)

    for data in dataloader:
        # set training = False to avoid norm + convert to BGR
        data_samples = data_preprocessor(data, False)['data_samples']

        real_key = 'gt_img' if metric.real_key is None else metric.real_key
        img = getattr(data_samples, real_key)

        real_feat_ = metric.extract_features(img)
        real_feat.append(real_feat_)

        if is_main_process():
            pbar.update(task, advance=len(real_feat_) * get_world_size())

    # stop the pbar
    if is_main_process():
        pbar.stop()

    # collect results
    real_feat = torch.cat(real_feat)
    # use `all_gather` here, gather tensor is much quicker than gather object.
    real_feat = all_gather(real_feat)

    # only cat on the main process
    if is_main_process():
        real_feat = torch.cat(real_feat, dim=0)[:len(dataloader.dataset)].cpu()
        if auto_save:
            vgg_state = dict(vgg_feat=real_feat, **args)
            with open(vgg_pkl, 'wb') as file:
                pickle.dump(vgg_state, file)
        return real_feat



# Copyright (c) OpenMMLab. All rights reserved.
from .lsgan import LSGAN
from .lsgan_discriminator import LSGANDiscriminator
from .lsgan_generator import LSGANGenerator

__all__ = ['LSGAN', 'LSGANDiscriminator', 'LSGANGenerator']



# Copyright (c) OpenMMLab. All rights reserved.
import warnings
from typing import Dict, List, Sequence, Union

import torch
from mmengine.evaluator import BaseMetric, Evaluator
from mmengine.runner.amp import autocast
from mmengine.runner.base_loop import BaseLoop
from torch.utils.data import DataLoader

from mmagic.registry import LOOPS
from .loop_utils import is_evaluator, update_and_check_evaluator

DATALOADER_TYPE = Union[DataLoader, Dict, List]
EVALUATOR_TYPE = Union[Evaluator, Dict, List]


@LOOPS.register_module()
class MultiValLoop(BaseLoop):
    """Validation loop for MMagic models which support evaluate multiply
    dataset at the same time. This class support evaluate:

    1. Metrics (metric) on a single dataset (e.g. PSNR and SSIM on DIV2K
       dataset)
    2. Different metrics on different datasets (e.g. PSNR on DIV2K and SSIM
       and PSNR on SET5)

    Use cases:

    Case 1: metrics on a single dataset

    >>> # add the following lines in your config
    >>> # 1. use `MultiValLoop` instead of `ValLoop` in MMEngine
    >>> val_cfg = dict(type='MultiValLoop')
    >>> # 2. specific MultiEvaluator instead of Evaluator in MMEngine
    >>> val_evaluator = dict(
    >>>     type='MultiEvaluator',
    >>>     metrics=[
    >>>         dict(type='PSNR', crop_border=2, prefix='Set5'),
    >>>         dict(type='SSIM', crop_border=2, prefix='Set5'),
    >>>     ])
    >>> # 3. define dataloader
    >>> val_dataloader = dict(...)

    Case 2: different metrics on different datasets

    >>> # add the following lines in your config
    >>> # 1. use `MultiValLoop` instead of `ValLoop` in MMEngine
    >>> val_cfg = dict(type='MultiValLoop')
    >>> # 2. specific a list MultiEvaluator
    >>> # do not forget to add prefix for each metric group
    >>> div2k_evaluator = dict(
    >>>     type='MultiEvaluator',
    >>>     metrics=dict(type='SSIM', crop_border=2, prefix='DIV2K'))
    >>> set5_evaluator = dict(
    >>>     type='MultiEvaluator',
    >>>     metrics=[
    >>>         dict(type='PSNR', crop_border=2, prefix='Set5'),
    >>>         dict(type='SSIM', crop_border=2, prefix='Set5'),
    >>>     ])
    >>> # define evaluator config
    >>> val_evaluator = [div2k_evaluator, set5_evaluator]
    >>> # 3. specific a list dataloader for each metric groups
    >>> div2k_dataloader = dict(...)
    >>> set5_dataloader = dict(...)
    >>> # define dataloader config
    >>> val_dataloader = [div2k_dataloader, set5_dataloader]

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or dict or list): A dataloader object or a dict
            to build a dataloader a list of dataloader object or a list of
            config dicts.
        evaluator (Evaluator or dict or list): A evaluator object or a dict to
            build the evaluator or a list of evaluator object or a list of
            config dicts.
    """

    def __init__(self,
                 runner,
                 dataloader: DATALOADER_TYPE,
                 evaluator: EVALUATOR_TYPE,
                 fp16: bool = False):
        self._runner = runner

        self.dataloaders = self._build_dataloaders(dataloader)
        self.evaluators = self._build_evaluators(evaluator)

        self.fp16 = fp16

        assert len(self.dataloaders) == len(self.evaluators), (
            'Length of dataloaders and evaluators must be same, but receive '
            f'\'{len(self.dataloaders)}\' and \'{len(self.evaluators)}\''
            'respectively.')

        self._total_length = None  # length for all dataloaders

    @property
    def total_length(self) -> int:
        if self._total_length is not None:
            return self._total_length

        warnings.warn('\'total_length\' has not been initialized and return '
                      '\'0\' for safety. This result is likely to be incorrect'
                      ' and we recommend you to call \'total_length\' after '
                      '\'self.run\' is called.')
        return 0

    def _build_dataloaders(self,
                           dataloader: DATALOADER_TYPE) -> List[DataLoader]:
        """Build dataloaders.

        Args:
            dataloader (Dataloader or dict or list): A dataloader object or a
                dict to build a dataloader a list of dataloader object or a
                list of config dict.

        Returns:
            List[Dataloader]: List of dataloaders for compute metrics.
        """
        runner = self._runner

        if not isinstance(dataloader, list):
            dataloader = [dataloader]

        dataloaders = []
        for loader in dataloader:
            if isinstance(loader, dict):
                dataloaders.append(
                    runner.build_dataloader(loader, seed=runner.seed))
            else:
                dataloaders.append(loader)

        return dataloaders

    def _build_evaluators(self, evaluator: EVALUATOR_TYPE) -> List[Evaluator]:
        """Build evaluators.

        Args:
            evaluator (Evaluator or dict or list): A evaluator object or a
                dict to build the evaluator or a list of evaluator object or a
                list of config dicts.

        Returns:
            List[Evaluator]: List of evaluators for compute metrics.
        """
        runner = self._runner

        # Input type checking and packing
        # 1. Single evaluator without type: [dict(), dict(), ...]
        # 2. Single evaluator with type: dict(type=xx, metrics=xx)
        # 3. Multi evaluator without type: [[dict, ...], [dict, ...]]
        # 4. Multi evaluator with type: [dict(type=xx, metrics=xx), dict(...)]
        if is_evaluator(evaluator):
            evaluator = [update_and_check_evaluator(evaluator)]
        else:
            assert all([
                is_evaluator(cfg) for cfg in evaluator
            ]), ('Unsupported evaluator type, please check your input and '
                 'the docstring.')
            evaluator = [update_and_check_evaluator(cfg) for cfg in evaluator]

        evaluators = [runner.build_evaluator(eval) for eval in evaluator]

        return evaluators

    def run(self):
        """Launch validation. The evaluation process consists of four steps.

        1. Prepare pre-calculated items for all metrics by calling
           :meth:`self.evaluator.prepare_metrics`.
        2. Get a list of metrics-sampler pair. Each pair contains a list of
           metrics with the same sampler mode and a shared sampler.
        3. Generate images for the each metrics group. Loop for elements in
           each sampler and feed to the model as input by calling
           :meth:`self.run_iter`.
        4. Evaluate all metrics by calling :meth:`self.evaluator.evaluate`.
        """

        self._runner.call_hook('before_val')
        self._runner.call_hook('before_val_epoch')
        self._runner.model.eval()

        # access to the true model
        module = self._runner.model
        if hasattr(self.runner.model, 'module'):
            module = module.module

        multi_metric = dict()
        idx_counter = 0
        self._total_length = 0

        # 1. prepare all metrics and get the total length
        metrics_sampler_lists = []
        meta_info_list = []
        dataset_name_list = []
        for evaluator, dataloader in zip(self.evaluators, self.dataloaders):
            # 1.1 prepare for metrics
            evaluator.prepare_metrics(module, dataloader)
            # 1.2 prepare for metric-sampler pair
            metrics_sampler_list = evaluator.prepare_samplers(
                module, dataloader)
            metrics_sampler_lists.append(metrics_sampler_list)
            # 1.3 update total length
            self._total_length += sum([
                len(metrics_sampler[1])
                for metrics_sampler in metrics_sampler_list
            ])
            # 1.4 save metainfo and dataset's name
            meta_info_list.append(
                getattr(dataloader.dataset, 'metainfo', None))
            dataset_name_list.append(dataloader.dataset.__class__.__name__)

        # 2. run evaluation
        for idx in range(len(self.evaluators)):
            # 2.1 set self.evaluator for run_iter
            self.evaluator = self.evaluators[idx]
            self.dataloader = self.dataloaders[idx]

            # 2.2 update metainfo for evaluator and visualizer
            meta_info = meta_info_list[idx]
            dataset_name = dataset_name_list[idx]
            if meta_info:
                self.evaluator.dataset_meta = meta_info
                self._runner.visualizer.dataset_meta = meta_info
            else:
                warnings.warn(
                    f'Dataset {dataset_name} has no metainfo. `dataset_meta` '
                    'in evaluator, metric and visualizer will be None.')

            # 2.3 generate images
            metrics_sampler_list = metrics_sampler_lists[idx]
            for metrics, sampler in metrics_sampler_list:
                for data in sampler:
                    self.run_iter(idx_counter, data, metrics)
                    idx_counter += 1

            # 2.4 evaluate metrics and update multi_metric
            metrics = self.evaluator.evaluate()
            if multi_metric and metrics.keys() & multi_metric.keys():
                raise ValueError('Please set different prefix for different'
                                 ' datasets in `val_evaluator`')
            else:
                multi_metric.update(metrics)

        # 3. finish evaluation and call hooks
        self._runner.call_hook('after_val_epoch', metrics=multi_metric)
        self._runner.call_hook('after_val')

    @torch.no_grad()
    def run_iter(self, idx, data_batch: dict, metrics: Sequence[BaseMetric]):
        """Iterate one mini-batch and feed the output to corresponding
        `metrics`.

        Args:
            idx (int): Current idx for the input data.
            data_batch (dict): Batch of data from dataloader.
            metrics (Sequence[BaseMetric]): Specific metrics to evaluate.
        """
        self._runner.call_hook(
            'before_val_iter', batch_idx=idx, data_batch=data_batch)
        # outputs should be sequence of BaseDataElement
        with autocast(enabled=self.fp16):
            outputs = self._runner.model.val_step(data_batch)
        self.evaluator.process(outputs, data_batch, metrics)
        self._runner.call_hook(
            'after_val_iter',
            batch_idx=idx,
            data_batch=data_batch,
            outputs=outputs)


@LOOPS.register_module()
class MultiTestLoop(BaseLoop):
    """Test loop for MMagic models which support evaluate multiply dataset at
    the same time. This class support evaluate:

    1. Metrics (metric) on a single dataset (e.g. PSNR and SSIM on DIV2K
       dataset)
    2. Different metrics on different datasets (e.g. PSNR on DIV2K and SSIM
       and PSNR on SET5)

    Use cases:

    Case 1: metrics on a single dataset

    >>> # add the following lines in your config
    >>> # 1. use `MultiTestLoop` instead of `TestLoop` in MMEngine
    >>> val_cfg = dict(type='MultiTestLoop')
    >>> # 2. specific MultiEvaluator instead of Evaluator in MMEngine
    >>> test_evaluator = dict(
    >>>     type='MultiEvaluator',
    >>>     metrics=[
    >>>         dict(type='PSNR', crop_border=2, prefix='Set5'),
    >>>         dict(type='SSIM', crop_border=2, prefix='Set5'),
    >>>     ])
    >>> # 3. define dataloader
    >>> test_dataloader = dict(...)

    Case 2: different metrics on different datasets

    >>> # add the following lines in your config
    >>> # 1. use `MultiTestLoop` instead of `TestLoop` in MMEngine
    >>> Test_cfg = dict(type='MultiTestLoop')
    >>> # 2. specific a list MultiEvaluator
    >>> # do not forget to add prefix for each metric group
    >>> div2k_evaluator = dict(
    >>>     type='MultiEvaluator',
    >>>     metrics=dict(type='SSIM', crop_border=2, prefix='DIV2K'))
    >>> set5_evaluator = dict(
    >>>     type='MultiEvaluator',
    >>>     metrics=[
    >>>         dict(type='PSNR', crop_border=2, prefix='Set5'),
    >>>         dict(type='SSIM', crop_border=2, prefix='Set5'),
    >>>     ])
    >>> # define evaluator config
    >>> test_evaluator = [div2k_evaluator, set5_evaluator]
    >>> # 3. specific a list dataloader for each metric groups
    >>> div2k_dataloader = dict(...)
    >>> set5_dataloader = dict(...)
    >>> # define dataloader config
    >>> test_dataloader = [div2k_dataloader, set5_dataloader]

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or dict or list): A dataloader object or a dict
            to build a dataloader a list of dataloader object or a list of
            config dicts.
        evaluator (Evaluator or dict or list): A evaluator object or a dict to
            build the evaluator or a list of evaluator object or a list of
            config dicts.
    """

    def __init__(self, runner, dataloader, evaluator, fp16=False):
        self._runner = runner

        self.dataloaders = self._build_dataloaders(dataloader)
        self.evaluators = self._build_evaluators(evaluator)

        self.fp16 = fp16

        assert len(self.dataloaders) == len(self.evaluators), (
            'Length of dataloaders and evaluators must be same, but receive '
            f'\'{len(self.dataloaders)}\' and \'{len(self.evaluators)}\''
            'respectively.')

        self._total_length = None

    @property
    def total_length(self) -> int:
        if self._total_length is not None:
            return self._total_length

        warnings.warn('\'total_length\' has not been initialized and return '
                      '\'0\' for safety. This result is likely to be incorrect'
                      ' and we recommend you to call \'total_length\' after '
                      '\'self.run\' is called.')
        return 0

    def _build_dataloaders(self,
                           dataloader: DATALOADER_TYPE) -> List[DataLoader]:
        """Build dataloaders.

        Args:
            dataloader (Dataloader or dict or list): A dataloader object or a
                dict to build a dataloader a list of dataloader object or a
                list of config dict.

        Returns:
            List[Dataloader]: List of dataloaders for compute metrics.
        """
        runner = self._runner

        if not isinstance(dataloader, list):
            dataloader = [dataloader]

        dataloaders = []
        for loader in dataloader:
            if isinstance(loader, dict):
                dataloaders.append(
                    runner.build_dataloader(loader, seed=runner.seed))
            else:
                dataloaders.append(loader)

        return dataloaders

    def _build_evaluators(self, evaluator: EVALUATOR_TYPE) -> List[Evaluator]:
        """Build evaluators.

        Args:
            evaluator (Evaluator or dict or list): A evaluator object or a
                dict to build the evaluator or a list of evaluator object or a
                list of config dicts.

        Returns:
            List[Evaluator]: List of evaluators for compute metrics.
        """
        runner = self._runner

        # Input type checking and packing
        # 1. Single evaluator without type: [dict(), dict(), ...]
        # 2. Single evaluator with type: dict(type=xx, metrics=xx)
        # 3. Multi evaluator without type: [[dict, ...], [dict, ...]]
        # 4. Multi evaluator with type: [dict(type=xx, metrics=xx), dict(...)]
        if is_evaluator(evaluator):
            evaluator = [update_and_check_evaluator(evaluator)]
        else:
            assert all([
                is_evaluator(cfg) for cfg in evaluator
            ]), ('Unsupported evaluator type, please check your input and '
                 'the docstring.')
            evaluator = [update_and_check_evaluator(cfg) for cfg in evaluator]

        evaluators = [runner.build_evaluator(eval) for eval in evaluator]

        return evaluators

    def run(self):
        """Launch validation. The evaluation process consists of four steps.

        1. Prepare pre-calculated items for all metrics by calling
           :meth:`self.evaluator.prepare_metrics`.
        2. Get a list of metrics-sampler pair. Each pair contains a list of
           metrics with the same sampler mode and a shared sampler.
        3. Generate images for the each metrics group. Loop for elements in
           each sampler and feed to the model as input by calling
           :meth:`self.run_iter`.
        4. Evaluate all metrics by calling :meth:`self.evaluator.evaluate`.
        """

        self._runner.call_hook('before_test')
        self._runner.call_hook('before_test_epoch')
        self._runner.model.eval()

        # access to the true model
        module = self._runner.model
        if hasattr(self._runner.model, 'module'):
            module = module.module

        multi_metric = dict()
        idx_counter = 0
        self._total_length = 0

        # 1. prepare all metrics and get the total length
        metrics_sampler_lists = []
        meta_info_list = []
        dataset_name_list = []
        for evaluator, dataloader in zip(self.evaluators, self.dataloaders):
            # 1.1 prepare for metrics
            evaluator.prepare_metrics(module, dataloader)
            # 1.2 prepare for metric-sampler pair
            metrics_sampler_list = evaluator.prepare_samplers(
                module, dataloader)
            metrics_sampler_lists.append(metrics_sampler_list)
            # 1.3 update total length
            self._total_length += sum([
                len(metrics_sampler[1])
                for metrics_sampler in metrics_sampler_list
            ])
            # 1.4 save metainfo and dataset's name
            meta_info_list.append(
                getattr(dataloader.dataset, 'metainfo', None))
            dataset_name_list.append(dataloader.dataset.__class__.__name__)

        # 2. run evaluation
        for idx in range(len(self.evaluators)):
            # 2.1 set self.evaluator for run_iter
            self.evaluator = self.evaluators[idx]
            self.dataloader = self.dataloaders[idx]

            # 2.2 update metainfo for evaluator and visualizer
            meta_info = meta_info_list[idx]
            dataset_name = dataset_name_list[idx]
            if meta_info:
                self.evaluator.dataset_meta = meta_info
                self._runner.visualizer.dataset_meta = meta_info
            else:
                warnings.warn(
                    f'Dataset {dataset_name} has no metainfo. `dataset_meta` '
                    'in evaluator, metric and visualizer will be None.')

            # 2.3 generate images
            metrics_sampler_list = metrics_sampler_lists[idx]
            for metrics, sampler in metrics_sampler_list:
                for data in sampler:
                    self.run_iter(idx_counter, data, metrics)
                    idx_counter += 1

            # 2.4 evaluate metrics and update multi_metric
            metrics = self.evaluator.evaluate()
            if multi_metric and metrics.keys() & multi_metric.keys():
                raise ValueError('Please set different prefix for different'
                                 ' datasets in `test_evaluator`')
            else:
                multi_metric.update(metrics)

        # 3. finish evaluation and call hooks
        self._runner.call_hook('after_test_epoch', metrics=multi_metric)
        self._runner.call_hook('after_test')

    @torch.no_grad()
    def run_iter(self, idx, data_batch: dict, metrics: Sequence[BaseMetric]):
        """Iterate one mini-batch and feed the output to corresponding
        `metrics`.

        Args:
            idx (int): Current idx for the input data.
            data_batch (dict): Batch of data from dataloader.
            metrics (Sequence[BaseMetric]): Specific metrics to evaluate.
        """
        self._runner.call_hook(
            'before_test_iter', batch_idx=idx, data_batch=data_batch)
        # outputs should be sequence of BaseDataElement
        with autocast(enabled=self.fp16):
            outputs = self._runner.model.test_step(data_batch)
        self.evaluator.process(outputs, data_batch, metrics)
        self._runner.call_hook(
            'after_test_iter',
            batch_idx=idx,
            data_batch=data_batch,
            outputs=outputs)
